        SUBROUTINE VODE_FSPLIT_VAR(N, T, Y, PR, DS)
        INCLUDE 'KPP_CRTFN.h'
        KPP_REAL Told, T
        KPP_REAL Y(NVAR), PR(NVAR), DS(NVAR)

        Told = TIME
        TIME = T
        CALL UPDATE_SUN()
        CALL UPDATE_RCONST()
        CALL FSPLIT_VAR( Y, RAD, PR, DS )
        TIME = Told
        END SUBROUTINE VODE_FSPLIT_VAR

        SUBROUTINE VODE_JACVAR_SP(N, T, Y, J)
        INCLUDE 'KPP_CRTFN.h'
        KPP_REAL Told, T
        KPP_REAL Y(NVAR), J(LU_NONZERO_V)

        Told = TIME
        TIME = T
        CALL UPDATE_SUN()
        CALL UPDATE_RCONST()
        CALL JACVAR_SP( Y, RAD, J )
        TIME = Told
        END SUBROUTINE VODE_JACVAR_SP


! -- STEM INTEGRATION WITH VODE.F --
        PROGRAM STEM
        INCLUDE 'KPP_CRTFN_s.h'

! --- PARAMETERS (FULL JACOBIAN)
        PARAMETER (ND=84,LRW=2*ND*ND+9*ND+25,LIW=ND+35)
        PARAMETER (LRCONT=4*ND+4+10)
        COMMON /CONT/ICONT(4),RCONT(LRCONT)
! --- DECLARATIONS
        DIMENSION RWORK(LRW),IWORK(LIW)
        COMMON/STAT/NFCN,NJAC,NSTEP,NACCPT,NREJCT,NDEC,NSOL
        COMMON /VERWER/ IVERWER, IBEGIN, STEPCUT
        EXTERNAL VODE_FSPLIT_VAR, VODE_JACVAR_SP
    
        write(6,*) 'ATMVODE. rtol=. stepcut=, DT=. DIRECT='
        read(5,*) rtols, stepcut, DT, idir
        
        STEPMIN = 0.
        STEPMAX = 10000.
        maxord = 5
        ibegin = 1
        
        
! --- DIMENSION OF THE SYSTEM
        N=ND

! ==== initial values ====
      CFACTOR = 8.12E+16
      cm = 1./CFACTOR
      TEMP = 241.42
      CALL  INITVAL ( )


! --- ENDPOINT OF INTEGRATION
       NDAYS = 1
       TSTART = 12.D0*3600.D0
       TEND = TSTART + 24.D0*3600.D0*NDAYS
! --- REQUIRED TOLERANCE, scalar ATOL (itol = 1), vector ATOL (itol = 2)
       atols=1e-18
       do i=1,NVAR
         ATOL(i) = atols
         RTOL(i) = rtols
       end do 
        ITOL=4
! ---- NORMAL COMPUTATION ---
        itask=1
        istate=1
! ---- USE OPTIONAL INPUT ---
        iopt=1   
        RWORK(6) = STEPMAX   ! STEP MAX
        RWORK(7) = STEPMIN   ! STEP MIN
        RWORK(5) = RWORK(7)  ! INITIAL STEP
        IWORK(5) = MAXORD    ! MAX ORD
        
        
        
! ----- SIGNAL FOR STIFF CASE, FULL JACOBIAN, INTERN (22) or SUPPLIED (21)
       MF = 21



! --- OUTPUT FILE ----
      OPEN(UNIT=10,FILE='Atmvode.m')
      write(10,*) 'atm=['


       TIME=TSTART

! --- PRINT  INITIAL CONDITIONS
!        WRITE (10,99) (TIME-tstart)/3600.D0,Var(I_O3)*CM,Var(I_HCHO)*CM, &
!           Var(I_H2O2)*CM,Var(I_NO)*CM,Var(I_NO2)*CM, &
!           Var(I_HNO3)*CM,Var(I_HNO4)*CM,Var(I_OH)*CM,Var(I_HO2)*CM


        DO WHILE (TIME.LE.TEND)

! --- CALL OF THE SUBROUTINE VODE
       if ( idir.eq.0 ) then
            ISTATE = 1
            IBEGIN = 1
       end if     
                    
       CALL dvode (VODE_FSPLIT_VAR, NVAR, Var, TIME, &
                   TIME+DT, itol, rtol, atol, &
                   itask, istate, iopt, rwork, lrw,  &
                   IWORK, liw, VODE_JACVAR_SP, mf,rpar,ipar)

       IF (ISTATE.LT.0) THEN
          WRITE(6,*) 'VODE WARNING:: ISTATE = ',ISTATE
       END IF   

! --- PRINT  SOLUTION
       if (time.le.tend) then


!        WRITE (10,99) (TIME-tstart)/3600.D0,Var(I_O3)*CM,Var(I_HCHO)*CM, &
!           Var(I_H2O2)*CM,Var(I_NO)*CM,Var(I_NO2)*CM, &
!           Var(I_HNO3)*CM,Var(I_HNO4)*CM,Var(I_OH)*CM,Var(I_HO2)*CM
!        percent=100.D0*(time-tstart)/(tend-tstart)
        write(6,95) percent,Var(I_O)*CM
        end if
 95     format(F7.2,' % DONE, O=',E10.4)       
 99     FORMAT(10(1X,E24.16))
 101    format(E24.16,E24.16,1X,I4,1X,E24.16)
         END DO
        write(6,*) 'O = ',Var(I_O)*CM 
        write(10,*) '];'
! --- PRINT STATISTICS
        WRITE (6,90) real(IWORK(21))/IWORK(11)*100., &
                    real(IWORK(22))/IWORK(11)*100.
 90     FORMAT(F9.3,' % conv, ',F9.3,'% err failures')
        CLOSE(10)
        STOP
        END
!
!
