      PROGRAM driver

      INCLUDE 'split.h'

      INTEGER i

      print *, 'QSSA for CBM-IV. RTOL=, Urban='  
      read(5,*) rtols, IS_URBAN 
      do i=1,NVAR
        RTOL(i) = rtols
        ATOL(i) = 1.0
      end do
      STEPMIN = 1.0d-2
      STEPMAX = 200.
      DT = 3600.

C ==== initial values ====
      CFACTOR = 2.55E+10
      CM = 1./CFACTOR
      TEMP = 288.15
      CALL INITVAL()
C --- Adjust initial conditions      
       IF (IS_URBAN.EQ.1) THEN
        do i=1,nvar
           Var(i) = 100.0
        end do   
        Var(I_NO)   = 50*CFACTOR 
        Var(I_NO2)  = 20*CFACTOR
        Var(I_HONO) = 1.0*CFACTOR
        Var(I_O3)   = 100.0*CFACTOR
        Var(I_H2O2) = 1.0*CFACTOR
        Var(I_CO)   = 300.0*CFACTOR
        Var(I_HCHO) = 10.0*CFACTOR
        Var(I_ALD2) = 10.0*CFACTOR
        Var(I_PAN)  = 1.0*CFACTOR
        Var(I_PAR)  = 50.0*CFACTOR 
        Var(I_OLE)  = 10.0*CFACTOR 
        Var(I_ETH)  = 10.0*CFACTOR
        Var(I_TOL)  = 2.0*CFACTOR 
        Var(I_XYL)  = 2.0*CFACTOR 
        Var(I_ISOP) = 10.0*CFACTOR
        Fix(I_H2O)  = 1.35E+7*CFACTOR
       ELSE
        do i=1,nvar
           Var(i) = 100.0
        end do   
        Var(I_NO)   = 0.1*CFACTOR 
        Var(I_NO2)  = 0.1*CFACTOR
        Var(I_O3)   = 30.0*CFACTOR
        Var(I_CO)   = 100.0*CFACTOR
        Var(I_H2O2) = 2.0*CFACTOR
        Var(I_ISOP) = 1.0*CFACTOR
        Fix(I_H2O)  = 1.35E+7*CFACTOR
       END IF
       CALL UPDATE_THERMAL()

      
c2      open(10, file='Qssa.m')
c2      write(10,992) (TIME-Tstart)/3600.D0,( Var(i), i=1,nvar)
       
      Tstart = 3600.*12.
      Tend = Tstart + 3600.*24.*5.

      
      
      DIV = 3600.0/DT

C ********** TIME LOOP *************************
      TIME = Tstart
      do while (TIME .lt. Tend)

C --- Add emissions
      IF (IS_URBAN.EQ.1) THEN
        Var(I_NO)   = Var(I_NO)   + 1.0*CFACTOR  /DIV
        Var(I_NO2)  = Var(I_NO2)  + 0.2*CFACTOR  /DIV  
        Var(I_CO)   = Var(I_CO)   + 2.0*CFACTOR  /DIV
        Var(I_HCHO) = Var(I_HCHO) + 0.2*CFACTOR  /DIV
        Var(I_ALD2) = Var(I_ALD2) + 0.2*CFACTOR  /DIV
        Var(I_PAR)  = Var(I_PAR)  + 2.0*CFACTOR  /DIV  
        Var(I_OLE)  = Var(I_OLE)  + 1.0*CFACTOR  /DIV 
        Var(I_ETH)  = Var(I_ETH)  + 0.2*CFACTOR  /DIV 
        Var(I_TOL)  = Var(I_TOL)  + 0.2*CFACTOR  /DIV 
        Var(I_XYL)  = Var(I_XYL)  + 0.2*CFACTOR  /DIV 
        Var(I_ISOP) = Var(I_ISOP) + 1.0*CFACTOR  /DIV
      ELSE
        VAR(I_NO)   = VAR(I_NO)   + 0.01*CFACTOR /DIV
        VAR(I_NO2)  = VAR(I_NO2)  + 0.01*CFACTOR /DIV
        VAR(I_ISOP) = VAR(I_ISOP) + 0.1*CFACTOR   /DIV
      END IF


        TITI = TIME
        TIME = TIME + DT/2.d0
        call UPDATE_SUN() 
        call UPDATE_PHOTO()
        TIME = TITI

        call INTEGRATE( DT )

        if (time.le.tend) then
          hours = (Time-Tstart)/3600.D0
          if ( dabs(hours-int(hours*10)/10) .lt. 0.01 ) then
c2            write(10,992) hours,  (Var(i), i=1,nvar)
          end if
        end if

      end do
C *********** END TIME LOOP ********

      close(10)
      STOP

991   FORMAT(10(X,E12.6))
992   FORMAT(100(X,E24.16))
      END

C End of MAIN function                                             
C **************************************************************** 


C -- QSSA WITH STEADY STATE APPROXIMATION --

      SUBROUTINE INTEGRATE( DT ) 

      INCLUDE 'split.h'

C DT - Integration interval
      REAL*8 DT

C Local variables
      REAL*8 P_VAR(NVAR), D_VAR(NVAR), V1(NVAR), V2(NVAR)
      LOGICAL IsReject
      REAL*8 T, Tnext, STEP, STEPold, Told, SUP
      REAL*8 ERR, ERRold, ratio, factor, facmax, tmp
      INTEGER I

      T = TIME
      Tnext = TIME + DT
      STEP = STEPMIN
      Told = T
      SUP  = 1e-14
      IsReject = .false.
      ERR = 1.d0
      ERRold = 1.d0


 10    continue  
       Tplus = T + STEP
       if ( Tplus .gt. Tnext ) then
          STEP = Tnext - T
          Tplus = Tnext
       end if


        call FSPLIT_VAR ( VAR, RAD, P_VAR, D_VAR )

        do i=1,NVAR
          IF ( d_var(i)*step .lt. 0.01) THEN       ! SLOW SPECIES
             XXX = STEP * (P_VAR(I) - D_VAR(I)*VAR(I))
             V1(I) = VAR(I) + XXX
             V2(I) = VAR(I) + 0.5*XXX
          ELSE IF ( d_var(i)*step .GT. 10.) THEN  ! FAST SPECIES
             V1(I) = P_VAR(I)/D_VAR(I)
             V2(I) = V1(I)
          ELSE                                    ! MEDIUM LIVED
            if ( abs(D_VAR(i)).gt.SUP ) then
              ratio = P_VAR(i)/D_VAR(i)
              tmp = exp(-D_VAR(i)*STEP*0.5)
              V1(i) = tmp * tmp * (VAR(i) - ratio) + ratio
              V2(i) = tmp * (VAR(i) - ratio) + ratio
            else
              tmp = D_VAR(i) * STEP * 0.5
              V1(i) = VAR(i) + P_VAR(i) * STEP * ( 1 - tmp *
     *             ( 1 - 2.0 / 3.0 * tmp ) )
              V2(i) = VAR(i) + P_VAR(i) * 0.5 * STEP*( 1-0.5*tmp*
     *             ( 1 - 1.0 / 3.0 * tmp ) )
              end if
          END IF
        end do

        CALL FSPLIT_VAR ( V2, RAD, P_VAR, D_VAR )

        do i=1,NVAR
          IF ( d_var(i)*step .lt. 0.01) THEN       ! SLOW SPECIES
             XXX = STEP * (P_VAR(I) - D_VAR(I)*VAR(I))
             V2(I) = V2(I) + 0.5*XXX
          ELSE IF ( d_var(i)*step .GT. 10.) THEN  ! FAST SPECIES
             V2(I) = P_VAR(I)/D_VAR(I)
          ELSE                                    ! MEDIUM LIVED
            if ( abs(D_VAR(i)).gt.SUP ) then
              ratio = P_VAR(i)/D_VAR(i)
              tmp = exp(-D_VAR(i)*STEP*0.5)
              V2(i) = tmp * (V2(i) - ratio) + ratio
            else
              tmp = D_VAR(i) * STEP * 0.5
              V2(i) = V2(i) + P_VAR(i) * 0.5 * STEP * ( 1 - 0.5 * tmp *
     *             ( 1 - 1.0 / 3.0 * tmp ) )
            end if
          END IF
        end do

C ===== Extrapolation and error estimation ========

        ERRold=ERR
        ERR=0.0D0
          do i=1,NVAR
             ERR = ERR + ((V2(i)-V1(i))/(ATOL(i) + RTOL(i)*V2(i)))**2
          end do       
        ERR = DSQRT( ERR/NVAR )
        STEPold=STEP

C ===== choosing the stepsize =====================

        factor = 0.9*ERR**(-0.35)*ERRold**0.2
        if (IsReject) then
            facmax=1.
        else
            facmax=8.
        end if 
        factor = DMAX1( 1.25D-1, DMIN1(factor,facmax) )
        STEP = DMIN1( STEPMAX, DMAX1(STEPMIN,factor*STEP) )    

C===================================================

        if ( (ERR.gt.1).and.(STEPold.gt.STEPMIN) ) then
          IsReject = .true.
        else
          IsReject = .false.
          do 140 i=1,NVAR
             Var(i)  = dmax1(V2(i), 0.d0)
 140      continue 
          T = Tplus     
        end if
      if ( T .lt. Tnext ) go to 10

      TIME = Tnext
      RETURN 
      END        
C End of INTEGRATE function                                        
C **************************************************************** 

