      SUBROUTINE INTEGRATE( TIN, TOUT )

      INCLUDE 'KPP_CRTFN.h'

C TIN - Start Time          
      KPP_REAL TIN
C TOUT - End Time
      KPP_REAL TOUT

        CALL ROS2(VAR,TIN,DT)

      TIME = TOUT  
  
      RETURN
      END

c ****************************************************************************
c Rosenbrock integrator routine

      SUBROUTINE ROS2(y,  t,  dtin)

      include 'KPP_CRTFN_s.h'

c Solve: d/dt c = f(t,c)
c
c written by Edwin Spee, CWI, Amsterdam.
c email: Edwin.Spee@cwi.nl
c if non-existent, try http://edwin-spee.mypage.org/
c last update: July 28, 1997
c
c on input:
c y      vector containing variable species at time=t
c fix    fixed species
c rconst reaction constants
c t      start time
c dtin   step size 
c
c on output:
c y      vector containing variable species at time = tend
c other parameters are not changed
c
c Uses datastructure and subroutine JACVAR_SP,DECOMP,F_VAR,SOLVE
c from KPP
c
c Integration method for Ros2:
c C_{n+1} = C_n + 3/2 dt k_1 + 1/2 dt k_2
c k_1 = S f(t_n, C_n)
c k_2 = S [f(t_{n+1},C_n + dt k_1) - 2 k_1]
c
c where g = 1.0 + sqrt(0.5),
c S = (I - g dt J ) ^ {-1}
c with J the Jacobian
c
      KPP_REAL y(nvar), k1(nvar), k2(nvar) ! real changed to KPP_REAL mz_rs_20010423
     +,w1(nvar)
     +,g,dtin,t
     +,J_chem(LU_NONZERO_V)
c     +,J_chem(LU_NONZERO_V+1)   so ists konsistent
     $,sxmin
      integer ier,it,nt,j,i

      g = 1.0 + sqrt(0.5)
      call JACVAR_SP(y,rad,J_chem)
      do j=1,LU_NONZERO_V
         J_chem(j) = -g*dtin*J_chem(j)
      enddo
crvk optionally cut this out and replace it by directly addressed statements
      do j=1,NVAR
         J_chem(lu_diag_v(j)) = J_chem(lu_diag_v(j)) + 1.0
      enddo

c***********************************************************************
crvk call to old routine with indirect adressing 
      call DECOMP (NVAR, J_chem, ier)
c new calls with direct addressing (needs to be preprocessed by write_decomp2d.f)
c uncomment if wanted
c additional input: sxmin = sxmin = sqrt(smallest representable number > 0
c     sxmin = 1.e-19 ! 4.e-154 for CRAY or double precision
      
c     call DECOMP_D (J_chem,ier,sxmin)  ! box model call
c     call DECOMP_2D (J_chem,ier,sxmin)  ! 2D call
c***********************************************************************
         
      if (ier.ne.0) then
         print *,'IER <> 0, dtin=',dtin
         stop
      endif
      
      call F_VAR(y,rad,k1)
      
      CALL SOLVE (J_chem,k1)
      
      do i = 1,nvar
         w1(i) = max(0.0, y(i) + dtin * k1(i) )
      enddo
      
      call F_VAR(w1,rad,k2)
      
      do i = 1,nvar
         k2(i) = k2(i) - 2.0 * k1(i)
      enddo
      CALL SOLVE (J_chem,k2)
      do i = 1,nvar
         y(i) = max(0.0,y(i) + 1.5*dtin * k1(i) + 0.5 *dtin*k2(i) )
      enddo
      end
