/***********************************************************************************
  KPP - symbolic chemistry Kinetics PreProcessor

  developed at Center for Global and Regional Atmospheric Research,
               University of Iowa

            by Valeriu Damian & Adrian Sandu 

***********************************************************************************/

#include "gdata.h"
#include "code.h"
#include "scan.h"

#define MAX_MONITOR 8
enum strutypes { PLAIN, LU };

int **structB;
int **structJ;
int **LUstructJ;

ICODE InlineCode[ INLINE_OPT ];

int NSPEC, NVAR, NVARACT, NRAD, NFIX, NREACT;
int NVARST, NRADST, NFIXST, PI; 
int C_DEFAULT, C;
int V, R, F, VAR, RAD, FIX;
int RCONST;
int A_VAR, P_VAR, D_VAR;
int A_RAD, P_RAD, D_RAD;
int KR, A, BV, BR, IV;
int JV, UV, JUV, JVS; 
int JR, UR, JUR, JRS;
int X, NTMPB;
int IROW_V, ICOL_V, CROW_V, DIAG_V;
int LU_IROW_V, LU_ICOL_V, LU_CROW_V, LU_DIAG_V, CNVAR, CNRAD;
int IROW_R, ICOL_R, CROW_R, DIAG_R;   
int LOOKAT, NLOOKAT, SLOOKAT, MONITOR, NMONITOR, SMONITOR;
int NMASS, SMASS;
int NONZERO_V, LU_NONZERO_V, NONZERO_R;
int TIME, SUN, TEMP;
int RTOLS, TSTART, TEND, DT;
int ATOL, RTOL, STEPMIN, STEPMAX, CFACTOR;
int NX, NY, NZ, MAX_XYZ, NS, TRANS, STRANS;
int U_VAR, U_RAD;

int nzv, LU_nzv, nzr;

NODE *sum, *prod;
int real;
int nlookat;
int nmoni;
int ntrans;
int nmass;
char * CommonName;

char * ascii(int x)
{
static char s[40];
  
  sprintf(s, "%d", x);
  return s;  
}

NODE * RConst( int n )
{
  switch( kr[n].type ) {
    case NUMBER:    return Const( kr[n].val.f );
    case PHOTO: 
    case EXPRESION: return Elm( RCONST, n );
  }
  return 0;
}

void InitGen()
{
int i,j;

  NSPEC   = DefConst( "NSPEC",   INT, "The number of species involved" );
  NVAR    = DefConst( "NVAR",    INT, "The number of Variable species" );
  NVARACT = DefConst( "NVARACT", INT, "The number of Active species" );
  NRAD    = DefConst( "NRAD",    INT, "The number of Radical species" );
  NFIX    = DefConst( "NFIX",    INT, "The number of Fixed species" );
  NREACT  = DefConst( "NREACT",  INT, "The number of reactions" );
  NVARST  = DefConst( "NVARST",  INT, "Starting of variables in conc. vect." );
  NRADST  = DefConst( "NRADST",  INT, "Starting of radicals in conc. vect." );
  NFIXST  = DefConst( "NFIXST",  INT, "Starting of fixed in conc. vect." );
  NONZERO_V = DefConst( "NONZERO_V", INT, "Number of nonzero variable elements" );
  LU_NONZERO_V = DefConst( "LU_NONZERO_V", INT, "Number of nonzero variable LU elements" );
  NONZERO_R = DefConst( "NONZERO_R", INT, "Number of nonzero radical elements" );
  CNVAR   = DefConst( "CNVAR",   INT, "(NVAR+1) The number of elements in compressed row format" );
  CNRAD   = DefConst( "CNRAD",   INT, "(NRAD+1) The number of elements in compressed row format" );

  PI      = DefConst( "PI",     real, "Value of pi" );

  VAR = DefvElm( "VAR", real, -NVAR, "Concentration of variable species (global)" );
  RAD = DefvElm( "RAD", real, -NRAD, "Concentration of radicals (global)" );
  FIX = DefvElm( "FIX", real, -NFIX, "Concentration of fix species (global)" );

  V = DefvElm( "V", real, -NVAR, "Concentration of variable species" );
  R = DefvElm( "R", real, -NRAD, "Concentration of radicals" );
  F = DefvElm( "F", real, -NFIX, "Concentration of fix species (global)" );

  U_VAR = DefvElm( "U_VAR", real, -NVAR, "Concentration of variable species in USER's order" );
  U_RAD = DefvElm( "U_RAD", real, -NVAR, "Concentration of radical species in USER's order" );
 
  RCONST = DefvElm( "RCONST", real, -NREACT, "Rate constants" );

  A_VAR = DefvElm( "A_VAR", real, -NVAR, "Agregate term" );
  P_VAR = DefvElm( "P_VAR", real, -NVAR, "Production term" );
  D_VAR = DefvElm( "D_VAR", real, -NVAR, "Destruction term" );

  A_RAD = DefvElm( "A_RAD", real, -NRAD, "Agregate term" );
  P_RAD = DefvElm( "P_RAD", real, -NRAD, "Production term" );
  D_RAD = DefvElm( "D_RAD", real, -NRAD, "Destruction term" );


  JVS   = DefvElm( "JVS", real, -LU_NONZERO_V, "the Jacobian of variables" );
  JRS   = DefvElm( "JRS", real, -NONZERO_R, "the Jacobian of radicals" );
    
  JV    = DefmElm( "JV", real, -NVAR, -NVAR, "the Jacobian of variables" );
  JR    = DefmElm( "JV", real, -NRAD, -NRAD, "the Jacobian of radicals" );
  UV    = DefvElm( "UV", real, -NVAR, "user vector for variables" );
  JUV   = DefvElm( "JUV", real, -NVAR, "Jacobian times user vector" );
  UR    = DefvElm( "UR", real, -NRAD, "user vector for radicals" );
  JUR   = DefvElm( "JUR", real, -NRAD, "Jacobian times user vector" );

  X    = DefvElm( "X", real, -NVAR, "vector for variables" );

  TIME  = DefElm( "TIME", real, "current integration time");
  SUN   = DefElm( "SUN", real, "light intensity");
  /* mz_rs_20021210: variable TEMP deleted */
  /* TEMP  = DefElm( "TEMP", real, "temperature"); */
  
  RTOLS  = DefElm( "RTOLS", real, "(scalar) relative tolerance");
  TSTART = DefElm( "TSTART", real, "integration start time");
  TEND   = DefElm( "TEND", real, "integration end time");
  DT     = DefElm( "DT", real, "integration step");
  
  A  = DefvElm( "A", real, -NREACT, "rate for each equation" );
  NTMPB   = DefConst( "NTMPB",   INT, "Length of Temporary Array B" );
  BV = DefvElm( "B", real, -NTMPB, "temporary matrix" );
  BR = DefmElm( "B", real, -NREACT, -NRAD, "temporary matrix" );
  KR = DefeElm( "KR", 0 );

  IROW_V  = DefvElm( "IROW_V", INT, -NONZERO_V, "row indexes for the Jacobian of variables" );
  ICOL_V  = DefvElm( "ICOL_V", INT, -NONZERO_V, "column indexes for the Jacobian of variables" );
  CROW_V  = DefvElm( "CROW_V", INT, -CNVAR, "compressed row indexes for the Jacobian of variables" );
  DIAG_V  = DefvElm( "DIAG_V", INT, -CNVAR, "diagonal indexes for the Jacobian of variables" );
  LU_IROW_V  = DefvElm( "LU_IROW_V", INT, -LU_NONZERO_V, "row indexes for the LU Jacobian of variables" );
  LU_ICOL_V  = DefvElm( "LU_ICOL_V", INT, -LU_NONZERO_V, "column indexes for the LU Jacobian of variables" );
  LU_CROW_V  = DefvElm( "LU_CROW_V", INT, -CNVAR, "compressed row indexes for the LU Jacobian of variables" );
  LU_DIAG_V  = DefvElm( "LU_DIAG_V", INT, -CNVAR, "diagonal indexes for the LU Jacobian of variables" );
  IROW_R  = DefvElm( "IROW_R", INT, -NONZERO_R, "row indexes for the Jacobian of radicals" );
  ICOL_R  = DefvElm( "ICOL_R", INT, -NONZERO_R, "column indexes for the Jacobian of radicals" );
  CROW_R  = DefvElm( "CROW_R", INT, -CNRAD, "compressed row indexes for the Jacobian of radicals" );
  DIAG_R  = DefvElm( "DIAG_R", INT, -CNRAD, "diagonal indexes for the Jacobian of radicals" );

  IV = DefeElm( "IV", 0 );

  C_DEFAULT = DefvElm( "C_DEFAULT", real, -NSPEC, "Default concentration for all species" );
  C = DefvElm( "C", real, -NSPEC, "Concentration for all species" );
  ATOL = DefvElm( "ATOL", real, -NSPEC, "Absolute tolerance" );
  RTOL = DefvElm( "RTOL", real, -NSPEC, "Relative tolerance" );

  STEPMIN  = DefElm( "STEPMIN", real, "minimum allowed intergation step");
  STEPMAX  = DefElm( "STEPMAX", real, "maximum allowed integration step");

  NLOOKAT = DefConst( "NLOOKAT", INT, "number of species to look at" );
  LOOKAT  = DefvElm( "LOOKAT", INT, -NLOOKAT, "indexes of species to look at" );
  SLOOKAT  = DefvElm( "SLOOKAT", STRING, -NLOOKAT, "names of species to look at" );

  NMONITOR = DefConst( "NMONITOR", INT, "number of species to monitor" );
  MONITOR  = DefvElm( "MONITOR", INT, -NMONITOR, "indexes of species to monitor" );
  SMONITOR  = DefvElm( "SMONITOR", STRING, -NMONITOR, "names of species to monitor" );

  NMASS = DefConst( "NMASS", INT, "number of atoms to check mass balance" );
  SMASS  = DefvElm( "SMASS", STRING, -NMASS, "names of atoms for mass balance" );

  NX = DefConst( "NX", INT, "X grid dimension" );
  NY = DefConst( "NY", INT, "Y grid dimension" );
  NZ = DefConst( "NZ", INT, "Z grid dimension" );
  MAX_XYZ = DefConst( "MAX_XYZ", INT, "maximum grid dimension" );
  NS = DefConst( "NS", INT, "number of species to transport" );
  TRANS  = DefvElm( "TRANS", INT, -NS, "indexes of species to transport" );
  STRANS  = DefvElm( "STRANS", STRING, -NS, "names of species to transport" );

  CFACTOR  = DefElm( "CFACTOR", real, "Conversion factor");

  for ( i=0; i<EqnNr; i++ ) 
    for ( j=0; j<SpcNr; j++ ) 
      structB[i][j] = ( cc[j][i] != 0 ) ? 1 : 0;
}

int NonZero( int stru, int start, int end, 
             int *row, int *col, int *crow, int *diag )
{
int nElm;
int i,j;
  
  nElm = 0;
  for (i = 0; i < end-start; i++) { 
    switch( useLang ) {
      case F95_LANG: crow[i] = 1+nElm;		/* F95 line */ 
                   break; 
      case F_LANG: crow[i] = 1+nElm;
                   break; 
      case C_LANG: crow[i] = nElm;
                   break; 
    }
    for (j = 0; j < end-start; j++) {
      if( (i == j) || ( (stru) ? LUstructJ[i+start][j+start] 
                               : structJ[i+start][j+start] ) ) {
        switch( useLang ) {
          case F95_LANG: row[nElm] = 1+i;	/* F95 line */ 
                       col[nElm] = 1+j;  
          case F_LANG: row[nElm] = 1+i;
                       col[nElm] = 1+j;  
                       break; 
          case C_LANG: row[nElm] = i;
                       col[nElm] = j;  
                       break; 
        }
        nElm++;
      }
      if( i == j ) {
        switch( useLang ) {
          case F95_LANG: diag[i] = 1+nElm-1;	/* F95 line */
                       break; 
          case F_LANG: diag[i] = 1+nElm-1;
                       break; 
          case C_LANG: diag[i] = nElm-1;
                       break; 
        }
      }  
    }
  }
  switch( useLang ) {
    case F95_LANG: crow[i] = 1+nElm;		/* F95 line *//*-1 here, this one works*/
                 diag[i] = 1+nElm;
                 break; 
    case F_LANG: crow[i] = 1+nElm;
                 diag[i] = 1+nElm;
                 break; 
    case C_LANG: crow[i] = nElm;
                 diag[i] = nElm; 
                 break; 
  }
  return nElm;
}

void GenerateGData()
{
int  i,j,k;
int  *crow;
int  *diag;
int  nElm;
int  *lookat;
char *slookat[MAX_SPECIES];
int  *moni;
char *smoni[MAX_SPECIES];
int  *trans;
char *strans[MAX_SPECIES];
char *smass[MAX_ATOMS];

  /* Allocate local data structures */
  if( (crow=(int*)malloc(MAX_SPECIES*sizeof(int))) == NULL)
     FatalError(-30,"Cannot allocate crow in GenerateGData");
  if( (diag=(int*)malloc(MAX_SPECIES*sizeof(int))) == NULL)
     FatalError(-30,"Cannot allocate diag in GenerateGData");
  if( (lookat=(int*)malloc(MAX_SPECIES*sizeof(int))) == NULL)
     FatalError(-30,"Cannot allocate lookat in GenerateGData");
  if( (moni=(int*)malloc(MAX_SPECIES*sizeof(int))) == NULL)
     FatalError(-30,"Cannot allocate monit in GenerateGData");
  if( (trans=(int*)malloc(MAX_SPECIES*sizeof(int))) == NULL)
     FatalError(-30,"Cannot allocate trans in GenerateGData");


  UseFile( srcFile );

  F_Inline(  "%6sBLOCK DATA GLOBAL_DATA\n", " " );

  F_Inline("      INCLUDE '%s.h'", rootFileName);

  /* mz_rs_20020426+ */
  /* F95_Inline(  "%6sBLOCK DATA GLOBAL_DATA\n", " " ); */  
  /* F95_Inline("      INCLUDE '%s.h'", rootFileName); */ /* F95 second */
  /* F95_Inline("      USE GDATA"); */
  /* mz_rs_20020426- */

  C_Inline("#include <stdio.h>");
  C_Inline("#include <stdlib.h>");
  C_Inline("#include <math.h>");
  C_Inline("#include \"%s.h\"", rootFileName);
  if( useJacSparse )
    C_Inline("#include \"%s_s.h\"", rootFileName);
  NewLines(1);

  GlobalDeclare( C_DEFAULT );

  GlobalDeclare( C );
  C_Inline("%s * %s = & %s[%s];", C_types[real], 
            varTable[VAR]->name, varTable[C]->name, varTable[NVARST]->name );
  C_Inline("%s * %s = & %s[%s];", C_types[real], 
            varTable[RAD]->name, varTable[C]->name, varTable[NRADST]->name );
  C_Inline("%s * %s = & %s[%s];", C_types[real], 
            varTable[FIX]->name, varTable[C]->name, varTable[NFIXST]->name );
  C_Inline("%s * %s = & %s[%s];", C_types[real], 
            varTable[F]->name, varTable[C]->name, varTable[NFIXST]->name );

  GlobalDeclare( RCONST );
  GlobalDeclare( TIME );
  GlobalDeclare( SUN );
  /* mz_rs_20021210: variable TEMP deleted */
  /*   GlobalDeclare( TEMP ); */
  GlobalDeclare( RTOLS );
  GlobalDeclare( TSTART );
  GlobalDeclare( TEND );
  GlobalDeclare( DT );
  GlobalDeclare( ATOL );
  GlobalDeclare( RTOL );
  GlobalDeclare( STEPMIN );
  GlobalDeclare( STEPMAX );
  
  if( useLang != F95_LANG )
    InitDeclare( CFACTOR, 0, (void*)&cfactor ); 	/* F95 second */
  
  NewLines(1);

  nlookat = 0;
  for (i = 0; i < SpcNr; i++)
    if ( SpeciesTable[Code[i]].lookat ) {
      switch( useLang ) {
        case F95_LANG: lookat[nlookat] = 1+i;		/* F95 line *//* -1 here */
                     break; 
        case F_LANG: lookat[nlookat] = 1+i;
                     break; 
        case C_LANG: lookat[nlookat] = i;
                     break; 
      }
      slookat[nlookat] = SpeciesTable[Code[i]].name;
      nlookat++;  
    }
  
  if( useLang != F95_LANG )     
    InitDeclare( LOOKAT, nlookat, (void*)lookat ); 	/* F95 second */

  if( useLang != F95_LANG )
    InitDeclare( SLOOKAT, nlookat, (void*)slookat );	/* F95 second */

  nmoni = 0;
  for (i = 0; i < SpcNr; i++)
    if ( SpeciesTable[Code[i]].moni ) {
      switch( useLang ) {
        case F95_LANG: moni[nmoni] = 1+i;		/* F95 line */
                     break; 
        case F_LANG: moni[nmoni] = 1+i;
                     break; 
        case C_LANG: moni[nmoni] = i;
                     break; 
      }
      smoni[nmoni] = SpeciesTable[Code[i]].name;
      nmoni++;
    }  

  if( nmoni > MAX_MONITOR ) {
    Warning( "%d species to monitorize. Too many, keeping %d.", 
             nmoni, MAX_MONITOR );
    nmoni = MAX_MONITOR;         
  }

  if( useLang != F95_LANG )    
    InitDeclare( MONITOR, nmoni, (void*)moni ); 	/* F95 second */

  if( useLang != F95_LANG )
    InitDeclare( SMONITOR, nmoni, (void*)smoni );	/* F95 second */

  ntrans = 0;
  for (i = 0; i < SpcNr; i++)
    if ( SpeciesTable[Code[i]].trans ) {
      switch( useLang ) {
        case F95_LANG: trans[ntrans] = 1+i;		/* F95 line */
                     break; 
        case F_LANG: trans[ntrans] = 1+i;
                     break; 
        case C_LANG: trans[ntrans] = i;
                     break; 
      }
      strans[ntrans] = SpeciesTable[Code[i]].name;
      ntrans++;  
    }
  
  if( useLang != F95_LANG )     
    InitDeclare( TRANS, ntrans, (void*)trans );		/* F95 second */

  if( useLang != F95_LANG )
    InitDeclare( STRANS, ntrans, (void*)strans );	/* F95 second */

  nmass = 0;
  for (i = 0; i < AtomNr; i++)
    if ( AtomTable[i].masscheck ) {
      smass[nmass] = AtomTable[i].name;
      nmass++;
    }  
  
  if( useLang != F95_LANG )
    InitDeclare( SMASS, nmass, (void*)smass );		/* F95 second */

  if( useLang != F95_LANG ) { /* mz_rs_20020426: all deactivated for f95 */
    NewLines(1);
    WriteComment("User defined global variables");
  
    switch( useLang ) {
    case C_LANG: bprintf( InlineCode[ C_DATA_INT ].code ); 
      bprintf( InlineCode[ C_DATA ].code ); 
      break;
    case F_LANG: bprintf( InlineCode[ F_DATA_INT ].code );	
      bprintf( InlineCode[ F_DATA ].code ); 
      break;
      /* case F95_LANG: bprintf( InlineCode[ F95_DATA_INT ].code ); */
      /* bprintf( InlineCode[ F95_DATA ].code ); */
      /* break; */
    }
    FlushBuf();

    NewLines(1);
    WriteComment("End user defined global variables");
    NewLines(1);
  }

  F_Inline( "%6sEND\n\n", " " );

  /* F95_Inline( "%6sEND\n\n", " " ); */  /* mz_rs_20020426 */

  /* Free local data structures */
  free(crow); free(diag); free(lookat); free(moni); free(trans);
}

void GenerateSparseData()
{
int* irow;
int* icol;
int* crow;
int* diag;
int nElm;

  if( generateSD == SD_OFF ) return;

/* Allocate local arrays */
if ( (irow = (int*)malloc(MAX_SPECIES*MAX_SPECIES*sizeof(int)))==NULL ) 
    FatalError(-30,"Cannot allocate irow in GenerateSparseData.");
if ( (icol = (int*)malloc(MAX_SPECIES*MAX_SPECIES*sizeof(int)))==NULL ) 
    FatalError(-30,"Cannot allocate icol in GenerateSparseData.");
if ( (crow = (int*)malloc(MAX_SPECIES*sizeof(int)))==NULL ) 
    FatalError(-30,"Cannot allocate crow in GenerateSparseData.");
if ( (diag = (int*)malloc(MAX_SPECIES*sizeof(int)))==NULL ) 
    FatalError(-30,"Cannot allocate diag in GenerateSparseData.");
    
  UseFile( srcFile );

  NewLines(1);
  
//  WriteComment("");
//  WriteComment("Sparse data");
//  WriteComment("");

  F_Inline(  "%6sBLOCK DATA SPARSE_DATA\n", " " );
  F_Inline("      INCLUDE '%s_s.h'", rootFileName);

//  F95_Inline(  "%6sBLOCK DATA SPARSE_DATA\n", " " ); 		/* F95 line */
//  F95_Inline("      INCLUDE '%s_s.h'", rootFileName);		/* F95 second */
//  F95_Inline("      USE GDATA");				/* F95 second */

  nzv = NonZero( PLAIN, 0, varNr, irow, icol, crow, diag );

  if(useLang == F95_LANG) 				/* F95 second */
	  goto F95_SparseData;

  if( generateSD & SD_IROW_V )
    InitDeclare( IROW_V, nzv, (void*)irow );
  if( generateSD & SD_ICOL_V )
    InitDeclare( ICOL_V, nzv, (void*)icol );
  if( generateSD & SD_CROW_V )
    InitDeclare( CROW_V, varNr+1, (void*)crow );
  if( generateSD & SD_DIAG_V )
    InitDeclare( DIAG_V, varNr+1, (void*)diag );

  LU_nzv = NonZero( LU, 0, varNr, irow, icol, crow, diag );
  if( generateSD & SD_LU_IROW_V )
  InitDeclare( LU_IROW_V, LU_nzv, (void*)irow );
  if( generateSD & SD_LU_ICOL_V )
  InitDeclare( LU_ICOL_V, LU_nzv, (void*)icol );
  if( generateSD & SD_LU_CROW_V )
  InitDeclare( LU_CROW_V, varNr+1, (void*)crow );
  if( generateSD & SD_LU_DIAG_V )
  InitDeclare( LU_DIAG_V, varNr+1, (void*)diag );

  nzr = NonZero( PLAIN, varNr, varNr+radNr, irow, icol, crow, diag );
  if( generateSD & SD_IROW_R )
    InitDeclare( IROW_R, nzr, (void*)irow );
  if( generateSD & SD_ICOL_R )
    InitDeclare( ICOL_R, nzr, (void*)icol );
  if( generateSD & SD_CROW_R )
    InitDeclare( CROW_R, radNr+1, (void*)crow );
  if( generateSD & SD_DIAG_R )
    InitDeclare( DIAG_R, radNr+1, (void*)diag );

  
F95_SparseData:
  NewLines(1);
  F_Inline( "%6sEND\n\n", " " );

//  F95_Inline( "%6sEND\n\n", " " ); 			/* F95 line */
  
  /* Free local arrays */
  free(irow); free(icol); free(crow); free(diag);
}

void GenerateSparseHeader()
{
  UseFile( sparseFile );

  F_Inline("      INCLUDE '%s.h'", rootFileName);

//  F95_Inline("      INCLUDE '%s.h'", rootFileName); 	/* F95 line */
  
  CommonName = "SDATA";

  WriteComment("");
  WriteComment("Sparse data");
  WriteComment("");

  ExternDeclare( IROW_V );
  ExternDeclare( ICOL_V );
  ExternDeclare( CROW_V );
  ExternDeclare( DIAG_V );

  ExternDeclare( LU_IROW_V );
  ExternDeclare( LU_ICOL_V );
  ExternDeclare( LU_CROW_V );
  ExternDeclare( LU_DIAG_V );

  ExternDeclare( IROW_R );
  ExternDeclare( ICOL_R );
  ExternDeclare( CROW_R );
  ExternDeclare( DIAG_R );
  
  NewLines(1);
}

void GenerateSparseModuleHeader()
{
int* irow;
int* icol;
int* crow;
int* diag;
int nElm;

  UseFile( sparseModuleFile );  

  if( generateSD == SD_OFF ) return;

if ( (irow = (int*)malloc(MAX_SPECIES*MAX_SPECIES*sizeof(int)))==NULL ) 
    FatalError(-30,"Cannot allocate irow in GenerateSparseData.");
if ( (icol = (int*)malloc(MAX_SPECIES*MAX_SPECIES*sizeof(int)))==NULL ) 
    FatalError(-30,"Cannot allocate icol in GenerateSparseData.");
if ( (crow = (int*)malloc(MAX_SPECIES*sizeof(int)))==NULL ) 
    FatalError(-30,"Cannot allocate crow in GenerateSparseData.");
if ( (diag = (int*)malloc(MAX_SPECIES*sizeof(int)))==NULL ) 
    FatalError(-30,"Cannot allocate diag in GenerateSparseData.");

  CommonName = "SDATA";

  F95_Inline("MODULE SPRSDATA");		 	/* F95 second */
  NewLines(1);

  /* mz_rs_20040202+ */
  //  F95_Inline("USE GDATA, ONLY: dp");
  //  NewLines(1);
  /* mz_rs_20040202- */

  /* mz_rs_20030728+ */
  F95_Inline("IMPLICIT NONE");
  NewLines(1);
  WriteComment("PUBLIC is already default");
  //  F95_Inline("PRIVATE :: dp"); /* mz_rs_20040202 */
  F95_Inline("SAVE");
  NewLines(1);
  /* mz_rs_20030728- */

  WriteComment("");
  WriteComment("Sparse data");
  WriteComment("");

  DeclareConstant( NSPEC,   ascii( SpcNr ) );
  DeclareConstant( NVAR,    ascii( max( varNr, 1 ) ) );
  DeclareConstant( NVARACT, ascii( max( varactNr, 1 ) ) );
  DeclareConstant( NRAD,    ascii( max( radNr, 1 ) ) );
  DeclareConstant( NFIX,    ascii( max( fixNr, 1 ) ) );
  DeclareConstant( NREACT,  ascii( EqnNr ) );
  DeclareConstant( NVARST,  ascii( varstNr ) );
  DeclareConstant( NRADST,  ascii( radstNr ) );
  DeclareConstant( NFIXST,  ascii( fixstNr ) );
  DeclareConstant( NONZERO_V,  ascii( nzv ) );

  LU_nzv = NonZero( LU, 0, varNr, irow, icol, crow, diag );

  DeclareConstant( LU_NONZERO_V,  ascii( LU_nzv ) );

  DeclareConstant( NONZERO_R,  ascii( nzr ) );
  DeclareConstant( CNVAR,   "NVAR + 1" );
  DeclareConstant( CNRAD,   "NRAD + 1" );


  nzv = NonZero( PLAIN, 0, varNr, irow, icol, crow, diag );
  
  if( generateSD & SD_IROW_V )
    SecondExtern( IROW_V, nzv, (void*)irow );
  else
    ExternDeclare( IROW_V );
	  
  if( generateSD & SD_ICOL_V )
    SecondExtern(ICOL_V, nzv, (void*)icol );
  else
    ExternDeclare( ICOL_V );
  
  if( generateSD & SD_CROW_V )
    SecondExtern( CROW_V, varNr+1, (void*)crow );
  else
    ExternDeclare( CROW_V );
  
  if( generateSD & SD_DIAG_V )
    SecondExtern( DIAG_V, varNr+1, (void*)diag );
  else
    ExternDeclare( DIAG_V );
  
  LU_nzv = NonZero( LU, 0, varNr, irow, icol, crow, diag );
  
  if( generateSD & SD_LU_IROW_V )
    SecondExtern( LU_IROW_V, LU_nzv, (void*)irow );
  else
    ExternDeclare( LU_IROW_V );
  
  if( generateSD & SD_LU_ICOL_V )
    SecondExtern( LU_ICOL_V, LU_nzv, (void*)icol );
  else 
    ExternDeclare( LU_ICOL_V );
  
  if( generateSD & SD_LU_CROW_V )
    SecondExtern( LU_CROW_V, varNr+1, (void*)crow );
  else
    ExternDeclare( LU_CROW_V );

  if( generateSD & SD_LU_DIAG_V )
    SecondExtern( LU_DIAG_V, varNr+1, (void*)diag );
  else
    ExternDeclare( LU_DIAG_V );
  
  nzr = NonZero( PLAIN, varNr, varNr+radNr, irow, icol, crow, diag );

  if( generateSD & SD_IROW_R )
    SecondExtern( IROW_R, nzr, (void*)irow );
  else
    ExternDeclare( IROW_R );
  
  if( generateSD & SD_ICOL_R )
    SecondExtern( ICOL_R, nzr, (void*)icol );
  else
    ExternDeclare( ICOL_R );
  
  if( generateSD & SD_CROW_R )
    SecondExtern( CROW_R, radNr+1, (void*)crow );
  else
    ExternDeclare( CROW_R );
  
  if( generateSD & SD_DIAG_R )
    SecondExtern( DIAG_R, radNr+1, (void*)diag );
  else
    ExternDeclare( DIAG_R );
  
  NewLines(1);

  F95_Inline( "END MODULE SPRSDATA" ); 			
  
  
  free(irow); free(icol); free(crow); free(diag);
}



void GenerateFuncVar()
{
int i, j, k;
int used;
int l, m;
int F_VAR, FSPLIT_VAR;

  if( varNr == 0 ) return;
  
  UseFile( srcFile ); 

  F_VAR      = DefFnc( "F_VAR",      3, "function for the derivatives of variables - Agregate form");
  FSPLIT_VAR = DefFnc( "FSPLIT_VAR", 4, "function for the derivatives of variables - Split form");

  if( useAggregate )
    FunctionBegin( F_VAR, V, R, A_VAR );
  else
    FunctionBegin( FSPLIT_VAR, V, R, P_VAR, D_VAR );

  NewLines(1);
  WriteComment("Local variables");
  Declare( A );
  
  NewLines(1);
  WriteComment("Computation of equation rates");
  
  for(j=0; j<EqnNr; j++) {
    used = 0;
    if( useAggregate ) {
      for (i = 0; i < varNr; i++) 
        if ( cf[i][j] != 0 ) { 
          used = 1;
          break;
        }
    } else {
      for (i = 0; i < varNr; i++) 
        if ( cd[i][j] != 0 ) { 
          used = 1;
	  break;
        }
    }  
    
    if ( used ) {    
      prod = RConst( j );
      for (i = 0; i < varNr; i++) 
        for (k = 1; k <= (int)cc[i][j]; k++ )
          prod = Mul( prod, Elm( V, i ) ); 
      for ( ; i < varNr + radNr; i++) 
        for (k = 1; k <= (int)cc[i][j]; k++ )
          prod = Mul( prod, Elm( R, i - varNr ) );
      for ( ; i < SpcNr; i++) 
        for (k = 1; k <= (int)cc[i][j]; k++ )
          prod = Mul( prod, Elm( F, i - varNr - radNr ) );
      Assign( Elm( A, j ), prod );
    }
  }

  if( useAggregate ) {

    NewLines(1);
    WriteComment("Aggregate function");

    for (i = 0; i < varNr; i++) {
      sum = Const(0);
      for (j = 0; j < EqnNr; j++) 
        sum = Add( sum, Mul( Const( cf[i][j] ), Elm( A, j ) ) );
      Assign( Elm( A_VAR, i ), sum );
    }    

  } else {
    
    NewLines(1);
    WriteComment("Production function");

    for (i = 0; i < varNr; i++) {
      sum = Const(0);
      for (j = 0; j < EqnNr; j++) 
        sum = Add( sum, Mul( Const( cd[i][j] ), Elm( A, j ) ) );
      Assign( Elm( P_VAR, i ), sum );
    }
    
    NewLines(1);
    WriteComment("Destruction function");

    for (i = 0; i < varNr; i++) {
      sum = Const(0);       
      for(j=0; j<EqnNr; j++) {
        if ( cc[i][j] == 0 ) continue;
        prod = Mul( RConst( j ), Const( cc[i][j] ) );
        for (l = 0; l < varNr; l++) {
          m=(int)cc[l][j] - (l==i);
          for (k = 1; k <= m; k++ )
            prod = Mul( prod, Elm( V, l ) );
        }     
        for ( ; l < varNr + radNr; l++) 
	  for (k = 1; k <= (int)cc[l][j]; k++ )
            prod = Mul( prod, Elm( R, l - varNr ) );
        for ( ; l < SpcNr; l++) 
          for (k = 1; k <= (int)cc[l][j]; k++ )
            prod = Mul( prod, Elm( F, l - varNr - radNr ) ); 
        sum = Add( sum, prod );
      }
      Assign( Elm( D_VAR, i ), sum );
    }
  }   

  if( useAggregate ) 
    FunctionEnd( F_VAR );
  else
    FunctionEnd( FSPLIT_VAR );
  
  FreeVariable( F_VAR );
  FreeVariable( FSPLIT_VAR );
}

void GenerateFuncRad()
{
int i, j, k;
int used;
int l, m;
int F_RAD, FSPLIT_RAD;

  if( radNr == 0 ) return;

  UseFile( srcFile );

  F_RAD      = DefFnc( "F_RAD",      3, "function for the derivatives of radicals - Agregate form");
  FSPLIT_RAD = DefFnc( "FSPLIT_RAD", 4, "function for the derivatives of radicals - Split form");
    
  if( useAggregate )
    FunctionBegin( F_RAD, V, R, A_RAD );
  else
    FunctionBegin( FSPLIT_RAD, V, R, P_RAD, D_RAD );
            
  NewLines(1);
  WriteComment("Local variables");
  Declare( A );
      
  NewLines(1);
  WriteComment("Computation of equation rates");
    
  for(j=0; j<EqnNr; j++) {
    used = 0;
    if( useAggregate ) {
      for (i = varNr; i < varNr + radNr; i++)
        if ( cf[i][j] != 0 ) { 
          used = 1;
          break;
        }
    } else {
      for (i = varNr; i < varNr + radNr; i++) 
        if ( cd[i][j] != 0 ) { 
          used = 1;
          break;
        }
    }  
    
    if ( used ) {    
      prod = RConst( j );
      for (i = 0; i < varNr; i++) 
        for (k = 1; k <= (int)cc[i][j]; k++ )
          prod = Mul( prod, Elm( V, i ) );
      for ( ; i < varNr + radNr; i++) 
        for (k = 1; k <= (int)cc[i][j]; k++ )
          prod = Mul( prod, Elm( R, i - varNr ) );  
      for ( ; i < SpcNr; i++) 
        for (k = 1; k <= (int)cc[i][j]; k++ )
          prod = Mul( prod, Elm( F, i - varNr - radNr ) );
      Assign( Elm( A, j ), prod );
    }
  }

  if( useAggregate ) {
    
    NewLines(1);
    WriteComment("Aggregate function");
    
    for (i = varNr; i < varNr + radNr; i++) {
      sum = Const(0);
      for (j = 0; j < EqnNr; j++)
        sum = Add( sum, Mul( Const( cf[i][j] ), Elm( A, j ) ) );
      Assign( Elm( A_RAD, i - varNr ), sum );
    }    

  } else {

    NewLines(1);
    WriteComment("Production function");

    for (i = varNr; i < varNr + radNr; i++) {
      sum = Const(0);
      for (j = 0; j < EqnNr; j++)
        sum = Add( sum, Mul( Const( cd[i][j] ), Elm( A, j ) ) ); 
      Assign( Elm( P_RAD, i - varNr ), sum ); 
    }
    
    NewLines(1);
    WriteComment("Destruction function");

    for (i = varNr; i < varNr + radNr; i++) {
      sum = Const(0);
      for(j=0; j<EqnNr; j++) {
        if ( cc[i][j] == 0 ) continue;
        prod = Mul( RConst( j ), Const( cc[i][j] ) );
        for (l = 0; l < varNr; l++)
          for (k = 1; k <= (int)cc[l][j]; k++ )
            prod = Mul( prod, Elm( V, l ) );
        for ( ; l < varNr + radNr; l++) { 
          m=(int)cc[l][j] - (l==i);
          for (k = 1; k <= m; k++ )
            prod = Mul( prod, Elm( R, l - varNr ) );
        }
        for ( ; l < SpcNr; l++) 
          for (k = 1; k <= (int)cc[l][j]; k++ )
            prod = Mul( prod, Elm( F, l - varNr - radNr ) );
        sum = Add( sum, prod );
      }
      Assign( Elm( D_RAD, i - varNr ), sum );
    }
  }   

  if( useAggregate )
    FunctionEnd( F_RAD );
  else
    FunctionEnd( FSPLIT_RAD );
  
  FreeVariable( F_RAD );
  FreeVariable( FSPLIT_RAD );
}

void GenerateJacVar()
{
int i,j,k,l,m;
int nElm, nonzeros_B;
int JACVAR_SP, JACVAR;
  
  if( varNr == 0 ) return;
  
  UseFile( srcFile );
  
  JACVAR_SP  = DefFnc( "JACVAR_SP", 3,
                  "function for the Jacobian of Variables using sparse matrix representation");
  JACVAR     = DefFnc( "JACVAR", 3, "function for the Jacobian of Variables");
  
  if( useJacSparse )
    FunctionBegin( JACVAR_SP, V, R, JVS );
  else
    FunctionBegin( JACVAR, V, R, JV );
  
  /* Each nonzero entry of B now counts its rank */
  nonzeros_B = 0;
  for ( i=0; i<EqnNr; i++ ) 
    for ( j=0; j<SpcNr; j++ ) 
       if ( structB[i][j] != 0 ) {
	 nonzeros_B++;
         structB[i][j] = nonzeros_B;
	 }
	 
  NewLines(1);
  WriteComment("Local variables");
  DeclareConstant( NTMPB,   ascii( nonzeros_B ) );
  Declare( BV );
        
  NewLines(1);

  for ( i=0; i<EqnNr; i++ ) {
    for ( j=0; j<varNr; j++ ) {
      if ( cc[j][i] != 0 ) {
        prod = Mul( RConst( i ), Const( cc[j][i] ) );
        for (l = 0; l < varNr; l++) { 
          m = (int)cc[l][i] - (l==j);
          for (k = 1; k <= m; k++ )
            prod = Mul( prod, Elm( V, l ) ); 
        }      
        for ( ; l < varNr + radNr; l++) 
          for (k = 1; k <= (int)cc[l][i]; k++ )
            prod = Mul( prod, Elm( R, l - varNr ) );
        for ( ; l < SpcNr; l++) 
          for (k = 1; k <= (int)cc[l][i]; k++ )
            prod = Mul( prod, Elm( F, l - varNr - radNr ) );
	/* Comment the B */
        switch( useLang ) {
           case C_LANG:  
	     WriteComment("B[%d] = dA[%d]/dV[%d]",structB[i][j]-1,i,j);
             break;
           case F_LANG: 
	     WriteComment("B(%d) = dA(%d)/dV(%d)",structB[i][j],i+1,j+1);	
             break;
           case F95_LANG: 
	     WriteComment("B(%d) = dA(%d)/dV(%d)",structB[i][j],i+1,j+1);	/* F95 line */
             break;
        }
        Assign( Elm( BV, structB[i][j]-1 ), prod );
      }
    }
  }

  nElm = 0;
  NewLines(1);
  WriteComment("Construct the Jacobian terms from B's"); 

  if ( useJacSparse ) {
  for (i = 0; i < varNr; i++) {
    for (j = 0; j < varNr; j++) {
      if( LUstructJ[i][j] ) {
        sum = Const(0);
        for (k = 0; k < EqnNr; k++) {
          if( cf[i][k]*structB[k][j] != 0 ) 
            sum = Add( sum, Mul( Const( cf[i][k] ), Elm( BV, structB[k][j]-1 ) ) );
        }
	/* Comment the B */
         switch( useLang ) {
           case C_LANG:  
	     WriteComment("JVS[%d] = JACVAR.FULL[%d][%d]",nElm,i,j);
             break;
           case F_LANG: 
	     WriteComment("JVS(%d) = JACVAR.FULL(%d,%d)",nElm+1,i+1,j+1);
             break;
           case F95_LANG: 
	     WriteComment("JVS(%d) = JACVAR.FULL(%d,%d)",nElm+1,i+1,j+1);	/* F95 line */
             break;
          }
          Assign( Elm( JVS, nElm ), sum );
          nElm++;
      } else {
        if( i == j ) {
          Assign( Elm( JVS, nElm ), Const(0) );
          nElm++;
        }
      }
    }
  }  
  } else { /* full Jacobian */
  for (i = 0; i < varNr; i++) {
    for (j = 0; j < varNr; j++) {
      if( structJ[i][j] ) {
        sum = Const(0);
        for (k = 0; k < EqnNr; k++) {
          if( cf[i][k]*structB[k][j] != 0 ) 
            sum = Add( sum, Mul( Const( cf[i][k] ), Elm( BV, structB[k][j]-1 ) ) );
        }
        Assign( Elm( JV, i, j ), sum );
      } 
    }
  }
  }  
  
  if( useJacSparse )
    FunctionEnd( JACVAR_SP );
  else
    FunctionEnd( JACVAR );

  FreeVariable( JACVAR_SP );
  FreeVariable( JACVAR );
} 

void GenerateJacRad()
{
int i,j,k,l,m;
int nElm;
int JACRAD_SP, JACRAD;
  
  if( radNr == 0 ) return;
  
  UseFile( srcFile );

  JACRAD_SP  = DefFnc( "JACRAD_SP", 3,
                  "function for the Jacobian of Radicals using sparse matrix representation");
  JACRAD     = DefFnc( "JACRAD", 3, "function for the Jacobian of Radicals");

  if( useJacSparse )
    FunctionBegin( JACRAD_SP, V, R, JRS );
  else
    FunctionBegin( JACRAD, V, R, JR );
  
  NewLines(1);
  WriteComment("Local variables");
  Declare( BR );
      
  NewLines(1);
  WriteComment("Computation of B(i,j) = dA(i)/dv(j)");

  for ( i = 0; i < EqnNr; i++ ) {
    for ( j = varNr; j < varNr + radNr; j++ ) {
      if ( cc[j][i] != 0 ) {
        prod = Mul( RConst( i ), Const( cc[j][i] ) );
        for (l = 0; l < varNr; l++) 
          for (k = 1; k <= (int)cc[l][i]; k++ )
            prod = Mul( prod, Elm( V, l ) );
        for ( ; l < varNr + radNr; l++) {
          m = (int)cc[l][i] - (l==j);
          for (k = 1; k <= m; k++ )
            prod = Mul( prod, Elm( R, l - varNr ) );
        }    
        for ( ; l < SpcNr; l++) 
          for (k = 1; k <= (int)cc[l][i]; k++ )
            prod = Mul( prod, Elm( F, l - varNr - radNr ) );
        Assign( Elm( BR, i, j - varNr ), prod ); 
      }
    }
  }

  nElm = 0;
  NewLines(1);
  WriteComment("Construct the Jacobian terms from B's");

  for (i = varNr; i < varNr + radNr; i++) {
    for (j = varNr; j < varNr + radNr; j++) {
      if( structJ[i][j] ) {
        sum = Const(0);
	for (k = 0; k < EqnNr; k++) {
          if( cf[i][k]*structB[k][j] != 0 ) 
            sum = Add( sum, Mul( Const( cf[i][k] ), Elm( BV, k, j - varNr ) ) );
        }
        if( useJacSparse ) {
          Assign( Elm( JRS, nElm ), sum );
          nElm++;
        }  
        else
          Assign( Elm( JR, i - varNr, j - varNr ), sum );
      } else {
        if( (i == j) && useJacSparse ) {
          Assign( Elm( JRS, nElm ), Const(0) );
          nElm++;
        }
      }
      
    }
  }  
  if( useJacSparse )
    FunctionEnd( JACRAD_SP );
  else
    FunctionEnd( JACRAD );

  FreeVariable( JACRAD_SP );
  FreeVariable( JACRAD );
}

void GenerateJVectVar()
{
int i, j, nElm;
int JACVAR_VEC;
int JACVAR_SP_VEC;

  if( varNr == 0 ) return;

  UseFile( srcFile );

  JACVAR_VEC = DefFnc( "JACVAR_VEC", 3, 
       "function for sparse multiplication: square Jacobian times vector");
  JACVAR_SP_VEC = DefFnc( "JACVAR_SP_VEC", 3, 
        "function for sparse multiplication: sparse Jacobian times vector");

  if ( useJacSparse ) {
    FunctionBegin( JACVAR_SP_VEC, JVS, UV, JUV ); 
    nElm = 0; 
    for( i = 0; i < varNr; i++) {
      sum = Const(0);
      for( j = 0; j < varNr; j++ )
        if( LUstructJ[i][j] ) {
          if( structJ[i][j] != 0 )
            sum = Add( sum, Mul( Elm( JVS, nElm ), Elm( UV, j ) ) ); 
	  nElm++;
	  }
      Assign( Elm( JUV, i ), sum );
    }  
    FunctionEnd( JACVAR_SP_VEC );
    }

  else {
    FunctionBegin( JACVAR_VEC, JV, UV, JUV );  
    for( i = 0; i < varNr; i++) {
      sum = Const(0);
      for( j = 0; j < varNr; j++ )
        if( structJ[i][j] != 0 )
          sum = Add( sum, Mul( Elm( JV, i, j ), Elm( UV, j ) ) ); 
      Assign( Elm( JUV, i ), sum );
    }  
    FunctionEnd( JACVAR_VEC );
  }

  FreeVariable( JACVAR_VEC );
  FreeVariable( JACVAR_SP_VEC );
}

void GenerateJVectRad()
{
int i, j;
int JACRAD_VEC;

  if( radNr == 0 ) return;

  UseFile( srcFile );

  JACRAD_VEC = DefFnc( "JACRAD_VEC", 3, "function for sparse multiplication");
  FunctionBegin( JACRAD_VEC, JR, UR, JUR );

  for( i = varNr; i < varNr + radNr; i++) {
    sum = Const(0);
    for( j = varNr; j < varNr + radNr; j++ )
      if( structJ[i][j] != 0 )
        sum = Add( sum, Mul( Elm( JR, i - varNr, j - varNr ), 
                             Elm( UR, j - varNr ) ) );
    Assign( Elm( JUR, i - varNr ), sum );
  }

  FunctionEnd( JACRAD_VEC );
  FreeVariable( JACRAD_VEC );
}

void GenerateSparseUtil()
{
int SUTIL;

  UseFile( srcFile );
    
  SUTIL = DefFnc( "SPARSE_UTIL", 0, "SPARSE Utility functions");  
  CommentFunctionBegin( SUTIL );

  IncludeCode( "%s/util/sutil", Home );  
  
  CommentFunctionEnd( SUTIL );
  FreeVariable( SUTIL );
}

void GenerateSolve()
{
int i, j;
int SOLVE;
int *irow;
int *icol;
int *crow;
int *diag;
int nElm;
int ibgn, iend;
int useLangOld;

/* Allocate local arrays */
if ( (irow = (int*)malloc(MAX_SPECIES*MAX_SPECIES*sizeof(int)))==NULL ) 
    FatalError(-30,"Cannot allocate irow in GenerateSolve.");
if ( (icol = (int*)malloc(MAX_SPECIES*MAX_SPECIES*sizeof(int)))==NULL ) 
    FatalError(-30,"Cannot allocate icol in GenerateSolve.");
if ( (crow = (int*)malloc(MAX_SPECIES*sizeof(int)))==NULL ) 
    FatalError(-30,"Cannot allocate crow in GenerateSolve.");
if ( (diag = (int*)malloc(MAX_SPECIES*sizeof(int)))==NULL ) 
    FatalError(-30,"Cannot allocate diag in GenerateSolve.");


  useLangOld = useLang;
  useLang = C_LANG;
  nElm = NonZero( LU, 0, varNr, irow, icol, crow, diag );
  useLang = useLangOld;
  
  UseFile( srcFile );

  SOLVE = DefFnc( "SOLVE", 2, "function for sparse back substitution");
  if ( useLang == F95_LANG )
      FunctionBeginSprs( SOLVE, JVS, X );
  else
      FunctionBegin( SOLVE, JVS, X );

  for( i = 0; i < varNr; i++) {
    ibgn = crow[i];
    iend = diag[i];
    if( ibgn <= iend ) {
      sum = Elm( X, i );
      if ( ibgn < iend ) { 
        for( j = ibgn; j < iend; j++ )
          sum = Sub( sum, Mul( Elm( JVS, j ), Elm( X, icol[j] ) ) );
        Assign( Elm( X, i ), sum );
      }
    }
  }

  for( i = varNr-1; i >=0; i--) {
    ibgn = diag[i] + 1;
    iend = crow[i+1]; 
    sum = Elm( X, i );
    for( j = ibgn; j < iend; j++ )
      sum = Sub( sum, Mul( Elm( JVS, j ), Elm( X, icol[j] ) ) );
    sum = Div( sum, Elm( JVS, diag[i] ) );
    Assign( Elm( X, i ), sum );
  }

  FunctionEnd( SOLVE );
  FreeVariable( SOLVE );
  
  /* Free Local Arrays */
  free(irow); free(icol); free(crow); free(diag);
}

void GenerateUpdateSun()
{
int UPDATE_SUN;

  UseFile( srcFile );
    
  UPDATE_SUN = DefFnc( "UPDATE_SUN", 0, "update SUN light using TIME");  
  CommentFunctionBegin( UPDATE_SUN );

  IncludeCode( "%s/util/UpdateSun", Home );
  
  CommentFunctionEnd( UPDATE_SUN );
  FreeVariable( UPDATE_SUN );
}

void GenerateUpdateRconst()
{
int i;
int UPDATE_RCONST;

  UseFile( srcFile );

  UPDATE_RCONST = DefFnc( "UPDATE_RCONST", 0, "function to update rate constants");
  
  FunctionBegin( UPDATE_RCONST );

  /* mz_rs_20030728+ */
  /* variable TEMP deleted */
  /* A and B cannot be used before they are defined; thus deleted */
  /* C_Inline("#define ARR(A,B) A * exp(B/TEMP)" ); */
  /* F_Inline("      %s ARR, A, B", F_types[real] ); */
  /* F_Inline("      ARR(A,B) = A * EXP(B/TEMP)" ); */
  /* F_Inline("      A = ARR(A,B)" ); */
  /* F95_Inline("      %s ARR, A, B", F95_types[real] ); */
  /* F95_Inline("      ARR(A,B) = A * EXP(B/TEMP)" ); */
  /* F95_Inline("      A = ARR(A,B)" );                  */
  /* mz_rs_20030728- */
    
  NewLines(1);

  /* mz_rs_20020426: new fragment type added */
  if( useLang == F95_LANG ) {
    WriteComment("User defined functions");
    bprintf( InlineCode[ F95_RCONST ].code );
    FlushBuf();
    NewLines(1);
  }
  
  for( i = 0; i < EqnNr; i++) {
    if( kr[i].type == EXPRESION )
      Assign( Elm( RCONST, i ), Elm( KR, kr[i].val.st ) );  
    if( kr[i].type == PHOTO )
      Assign( Elm( RCONST, i ), Elm( KR, kr[i].val.st ) );  
  }
  
  FunctionEnd( UPDATE_RCONST );
  FreeVariable( UPDATE_RCONST );
}

void GenerateUpdatePhoto()
{
int i;
int UPDATE_PHOTO;

  UseFile( srcFile );

  UPDATE_PHOTO = DefFnc( "UPDATE_PHOTO", 0, "function to update photolitical rate constants");
  
  FunctionBegin( UPDATE_PHOTO );
  
  /* mz_rs_20030728+ */
  /* variable TEMP deleted */
  /* A and B cannot be used before they are defined; thus deleted */
  /* C_Inline("#define ARR(A,B) A * exp(B/TEMP)" ); */
  /* F_Inline("      %s ARR, A, B", F_types[real] ); */
  /* F_Inline("      ARR(A,B) = A * EXP(B/TEMP)" ); */
  /* F_Inline("      A = ARR(A,B)" ); */
  /* F95_Inline("      %s ARR, A, B", F95_types[real] ); */
  /* F95_Inline("      ARR(A,B) = A * EXP(B/TEMP)" ); */
  /* F95_Inline("      A = ARR(A,B)" );                  */
  /* mz_rs_20030728- */
  
  NewLines(1);
  
  /* mz_rs_20020426: new fragment type added */
  if( useLang == F95_LANG ) {
    WriteComment("User defined functions");
    bprintf( InlineCode[ F95_RCONST ].code );
    FlushBuf();
    NewLines(1);
  }
  
  for( i = 0; i < EqnNr; i++) {
    if( kr[i].type == PHOTO )
      Assign( Elm( RCONST, i ), Elm( KR, kr[i].val.st ) );  
  }
  
  FunctionEnd( UPDATE_PHOTO );
  FreeVariable( UPDATE_PHOTO );
}

void GenerateUpdateC()
{
int UPDATE_C;

  UseFile( srcFile );
    
  UPDATE_C = DefFnc( "UPDATE_C", 0, "update concentrations for transport");  
  FunctionBegin( UPDATE_C );

/* ... */

  FunctionEnd( UPDATE_C );
  FreeVariable( UPDATE_C );
}

void GenerateIntegrate()
{
int TIN, TOUT, INTEGRATE;

  UseFile( srcFile );
    

  TIN = DefElm( "TIN", real, "Start Time for Integration");
  TOUT = DefElm( "TOUT", real, "End Time for Integration");
  INTEGRATE = DefFnc( "INTEGRATE", 2, "Integrator routine");  
  CommentFunctionBegin( INTEGRATE, TIN, TOUT );

  if( strchr( integrator, '/' ) )
    IncludeCode( integrator );
  else 
    IncludeCode( "%s/int/%s", Home, integrator );  
  
  CommentFunctionEnd( INTEGRATE );
  FreeVariable( INTEGRATE );
}

void GenerateDriver()
{
int MAIN;

  UseFile( srcFile );

  NewLines(1);
  WriteComment("User defined functions");
  
  switch( useLang ) {
    case C_LANG: bprintf( InlineCode[ C_UTIL_INT ].code ); 
                 bprintf( InlineCode[ C_UTIL ].code ); 
                 break;
    case F_LANG: bprintf( InlineCode[ F_UTIL_INT ].code );
                 bprintf( InlineCode[ F_UTIL ].code ); 
                 break;
    case F95_LANG: bprintf( InlineCode[ F95_UTIL_INT ].code );		/* F95 line */
                 bprintf( InlineCode[ F95_UTIL ].code ); 
                 break;
  }
  FlushBuf();

  NewLines(1);
  WriteComment("End user defined functions");
  NewLines(1);
    
  MAIN = DefFnc( "MAIN", 0, "Main program - driver routine");  
  CommentFunctionBegin( MAIN );

  if( strchr( driver, '/' ) )
    IncludeCode( driver );
  else 
    IncludeCode( "%s/drv/%s", Home, driver );  
  
  CommentFunctionEnd( MAIN );
  FreeVariable( MAIN );
}

void GenerateUtil()
{
int UTIL;

  UseFile( srcFile );
    
  UTIL = DefFnc( "UTIL", 0, "Utility functions");  
  CommentFunctionBegin( UTIL);

  IncludeCode( "%s/util/util", Home );  
  
  CommentFunctionEnd( UTIL );
  FreeVariable( UTIL );
}

void GenerateHeader()
{
int spc;
int i;

int  *crow;
int  *diag;
int  nElm;

char name[20];
int offs;
int mxyz;

int  *lookat;
char *slookat[MAX_SPECIES];
int  *moni;
char *smoni[MAX_SPECIES];
int  *trans;
char *strans[MAX_SPECIES];
char *smass[MAX_ATOMS];

int j,dummy_species; /* mz_rs_20020914 */

  UseFile( ModuleHeadFile );
  
  CommonName = "GDATA";
  
  F_Inline("      IMPLICIT %s ( A-H, O-Z )", F_types[ real ] );

  F95_Inline("MODULE GDATA");		 	/* F95 second */
  NewLines(1);

  F95_Inline("IMPLICIT NONE"); /* mz_rs_20030728 */
  NewLines(1);

  F95_Inline("SAVE"); /* mz_rs_20030606 */
  NewLines(1);

  /* mz_rs_20040202+ */
  WriteComment("dp - definition of double precision");
  F95_Inline("      INTEGER, PARAMETER :: dp = SELECTED_REAL_KIND(12,307)");
  NewLines(1);
  /* mz_rs_20040202- */

  DeclareConstant( NSPEC,   ascii( SpcNr ) );
  DeclareConstant( NVAR,    ascii( max( varNr, 1 ) ) );
  DeclareConstant( NVARACT, ascii( max( varactNr, 1 ) ) );
  DeclareConstant( NRAD,    ascii( max( radNr, 1 ) ) );
  DeclareConstant( NFIX,    ascii( max( fixNr, 1 ) ) );
  DeclareConstant( NREACT,  ascii( EqnNr ) );
  DeclareConstant( NVARST,  ascii( varstNr ) );
  DeclareConstant( NRADST,  ascii( radstNr ) );
  DeclareConstant( NFIXST,  ascii( fixstNr ) );

  DeclareConstant( NONZERO_V,  ascii( nzv ) );
//  LU_nzv = NonZero( LU, 0, varNr, irow, icol, crow, diag );
  DeclareConstant( LU_NONZERO_V,  ascii( LU_nzv ) );
  DeclareConstant( NONZERO_R,  ascii( nzr ) );
  DeclareConstant( CNVAR,   "NVAR + 1" );
  DeclareConstant( CNRAD,   "NRAD + 1" );


  DeclareConstant( NLOOKAT,  ascii( nlookat ) );
  DeclareConstant( NMONITOR,  ascii( nmoni ) );
  DeclareConstant( NMASS, ascii( nmass ) );
  DeclareConstant( NX,  ascii( xNr ) );
  DeclareConstant( NY,  ascii( yNr ) );
  DeclareConstant( NZ,  ascii( zNr ) );
  mxyz = (xNr > yNr) ? xNr : yNr;
  mxyz = (mxyz > zNr) ? mxyz : zNr;
  DeclareConstant( MAX_XYZ,  ascii( mxyz ) );
  DeclareConstant( NS,  ascii( ntrans ) );

  DeclareConstant( PI, "3.14159265358979" );

  NewLines(1);
  WriteComment("Declaration for global variables");
  NewLines(1);

  NewLines(1);

  ExternDeclare( C_DEFAULT );

  ExternDeclare( C );
  
  /* here need change */ 
  if( useLang == F_LANG ) {
    Declare( VAR );
    Declare( RAD );
    Declare( FIX );
    Declare( F );
    F_Inline("      					/* F95 second */EQUIVALENCE( %s(1+%s),%s(1) )", 
            varTable[C]->name, varTable[NVARST]->name, varTable[VAR]->name );
    F_Inline("      EQUIVALENCE( %s(1+%s),%s(1) )", 
            varTable[C]->name, varTable[NRADST]->name, varTable[RAD]->name );
    F_Inline("      EQUIVALENCE( %s(1+%s),%s(1) )", 
            varTable[C]->name, varTable[NFIXST]->name, varTable[FIX]->name );
    F_Inline("      EQUIVALENCE( %s(1+%s),%s(1) )", 
            varTable[C]->name, varTable[NFIXST]->name, varTable[F]->name );
  }

  if( useLang == F95_LANG ) {
    Declare( VAR );
    Declare( RAD );
    Declare( FIX );
    Declare( F );
    F95_Inline("      EQUIVALENCE( %s(1+%s),%s(1) )", 		/* F95 line */
            varTable[C]->name, varTable[NVARST]->name, varTable[VAR]->name );
    F95_Inline("      EQUIVALENCE( %s(1+%s),%s(1) )", 
            varTable[C]->name, varTable[NRADST]->name, varTable[RAD]->name );
    F95_Inline("      EQUIVALENCE( %s(1+%s),%s(1) )", 
            varTable[C]->name, varTable[NFIXST]->name, varTable[FIX]->name );
    F95_Inline("      EQUIVALENCE( %s(1+%s),%s(1) )", 
            varTable[C]->name, varTable[NFIXST]->name, varTable[F]->name );
  }

  C_Inline("  extern %s * %s;", C_types[real], varTable[VAR]->name );
  C_Inline("  extern %s * %s;", C_types[real], varTable[RAD]->name );
  C_Inline("  extern %s * %s;", C_types[real], varTable[FIX]->name );
  C_Inline("  extern %s * %s;", C_types[real], varTable[F]->name );


  ExternDeclare( RCONST );
  ExternDeclare( TIME );
  ExternDeclare( SUN );
  /* mz_rs_20021210: variable TEMP deleted */
  /* ExternDeclare( TEMP ); */
  ExternDeclare( RTOLS );
  ExternDeclare( TSTART );
  ExternDeclare( TEND );
  ExternDeclare( DT );
  ExternDeclare( ATOL );
  ExternDeclare( RTOL );
  ExternDeclare( STEPMIN );
  ExternDeclare( STEPMAX );

  SecondExtern( CFACTOR, 0, (void*)&cfactor );
  nlookat = 0;
  for (i = 0; i < SpcNr; i++)
    if ( SpeciesTable[Code[i]].lookat ) {
      switch( useLang ) {
        case F95_LANG: lookat[nlookat] = 1+i;		
                     break; 
        case F_LANG: lookat[nlookat] = 1+i;
                     break; 
        case C_LANG: lookat[nlookat] = i;
                     break; 
      }
      slookat[nlookat] = SpeciesTable[Code[i]].name;
      nlookat++;  
    }
  SecondExtern( LOOKAT, nlookat, (void*)lookat ); 
  SecondExtern( SLOOKAT, nlookat, (void*)slookat );	/* F95 second */

  nmoni = 0;
  for (i = 0; i < SpcNr; i++)
    if ( SpeciesTable[Code[i]].moni ) {
      switch( useLang ) {
        case F95_LANG: moni[nmoni] = 1+i;		/* F95 line */
                     break; 
        case F_LANG: moni[nmoni] = 1+i;
                     break; 
        case C_LANG: moni[nmoni] = i;
                     break; 
      }
      smoni[nmoni] = SpeciesTable[Code[i]].name;
      nmoni++;
    }  

  SecondExtern( MONITOR, nmoni, (void*)moni );
  SecondExtern( SMONITOR, nmoni, (void*)smoni );

  ntrans = 0;
  for (i = 0; i < SpcNr; i++)
    if ( SpeciesTable[Code[i]].trans ) {
      switch( useLang ) {
        case F95_LANG: trans[ntrans] = 1+i;		/* F95 line */
                     break; 
        case F_LANG: trans[ntrans] = 1+i;
                     break; 
        case C_LANG: trans[ntrans] = i;
                     break; 
      }
      strans[ntrans] = SpeciesTable[Code[i]].name;
      ntrans++;  
    }

  
//    SecondExtern( TRANS, ntrans, (void*)trans );	

//    SecondExtern( STRANS, ntrans, (void*)strans );	
  ExternDeclare( TRANS );
  ExternDeclare( STRANS );
  
  nmass = 0;
  for (i = 0; i < AtomNr; i++)
    if ( AtomTable[i].masscheck ) {
      smass[nmass] = AtomTable[i].name;
      nmass++;
    }  
  
  SecondExtern( SMASS, nmass, (void*)smass ); 


  switch( useLang ) {
    case F95_LANG: offs = 1; break;		/* F95 line */
    case F_LANG: offs = 1; break;
    case C_LANG: offs = 0; break;
  }

  NewLines(1);
  WriteComment("Indexes declaration for variable species");
  NewLines(1);
  for( i = 0; i < varNr; i++) {
    sprintf( name, "I_%s", SpeciesTable[ Code[i] ].name );
    spc = DefConst( name, INT, 0 );
    DeclareConstant( spc, ascii( i+offs ) );
    FreeVariable( spc );
  }
 
  NewLines(1);
  WriteComment("Indexes declaration for radical species");
  NewLines(1);
  for( i = 0; i < radNr; i++) {
    sprintf( name, "I_%s", SpeciesTable[ Code[i + varNr] ].name );
    spc = DefConst( name, INT, 0 );
    DeclareConstant( spc, ascii( i+offs ) );
    FreeVariable( spc );
  }
 
  NewLines(1);
  WriteComment("Indexes declaration for fixed species");
  NewLines(1);
  for( i = 0; i < fixNr; i++) {
    sprintf( name, "I_%s", SpeciesTable[ Code[i + varNr + radNr] ].name );
    spc = DefConst( name, INT, 0 );
    DeclareConstant( spc, ascii( i+offs ) );
    FreeVariable( spc );
  }

  /* mz_rs_20020914+ set I_* to 0 for dummy species */
  NewLines(1);
  WriteComment("Indexes declaration for dummy species");
  NewLines(1);
  for( i = 0; i < MAX_SPECIES; i++) {
    if (SpeciesTable[i].type == 0) continue;
    dummy_species = 1;
    for( j = 0; j < MAX_SPECIES; j++) 
      if (Code[j] == i) dummy_species = 0;
    if (dummy_species) {    
      sprintf( name, "I_%s", SpeciesTable[i].name );
      spc = DefConst( name, INT, 0 );
      DeclareConstant( spc, ascii( 0 ) );
      FreeVariable( spc );
    }
  }
  /* mz_rs_20020914- */

  /* mz_rs_20021023+ */
  /* and now the whole block for KPP_ instead of I_ */
  /* with indices for C() instead of indices for VAR(), RAD(), and FIX()*/

  NewLines(1);
  WriteComment("Indexes KPP_ for variable species");
  NewLines(1);
  for( i = 0; i < varNr; i++) {
    sprintf( name, "KPP_%s", SpeciesTable[ Code[i] ].name );
    spc = DefConst( name, INT, 0 );
    DeclareConstant( spc, ascii( i+offs ) );
    FreeVariable( spc );
  }
 
  NewLines(1);
  WriteComment("Indexes KPP_ for radical species");
  NewLines(1);
  for( i = 0; i < radNr; i++) {
    sprintf( name, "KPP_%s", SpeciesTable[ Code[i + varNr] ].name );
    spc = DefConst( name, INT, 0 );
    DeclareConstant( spc, ascii( i+varNr+offs ) );
    FreeVariable( spc );
  }
 
  NewLines(1);
  WriteComment("Indexes KPP_ for fixed species");
  NewLines(1);
  for( i = 0; i < fixNr; i++) {
    sprintf( name, "KPP_%s", SpeciesTable[ Code[i + varNr + radNr] ].name );
    spc = DefConst( name, INT, 0 );
    DeclareConstant( spc, ascii( i+varNr+radNr+offs ) );
    FreeVariable( spc );
  }

  NewLines(1);
  WriteComment("Indexes KPP_ for dummy species");
  NewLines(1);
  for( i = 0; i < MAX_SPECIES; i++) {
    if (SpeciesTable[i].type == 0) continue;
    dummy_species = 1;
    for( j = 0; j < MAX_SPECIES; j++) 
      if (Code[j] == i) dummy_species = 0;
    if (dummy_species) {    
      sprintf( name, "KPP_%s", SpeciesTable[i].name );
      spc = DefConst( name, INT, 0 );
      DeclareConstant( spc, ascii( 0 ) );
      FreeVariable( spc );
    }
  }
  /* mz_rs_20021023- */

  NewLines(1);
  WriteComment("User defined variables");
  
  switch( useLang ) {
    case C_LANG: bprintf( InlineCode[ C_DECL_INT ].code ); 
                 bprintf( InlineCode[ C_DECL ].code ); 
                 break;
    case F_LANG: bprintf( InlineCode[ F_DECL_INT ].code );
                 bprintf( InlineCode[ F_DECL ].code );
                 break;
    case F95_LANG: bprintf( InlineCode[ F95_DECL_INT ].code );	/* F95 line */
                 bprintf( InlineCode[ F95_DECL ].code );
                 break;
  }
  FlushBuf();

  NewLines(1);
  WriteComment("End user defined variables");
  NewLines(1);
  F95_Inline("END MODULE GDATA"); /* F95 second */
  NewLines(1);

}

void WriteSpec( int i, int j )
{
char buf[100];

  if( Reactive[j] )
    sprintf( buf, "%s (r)", SpeciesTable[ Code[j] ].name );
  else  
    sprintf( buf, "%s (n)", SpeciesTable[ Code[j] ].name );
  WriteAll("%3d = %-10s", 1 + i, buf );
  FlushBuf();
}

void GenerateMap()
{
int i;
int dn;

  UseFile( mapFile );

  WriteAll("### Options -------------------------------------------\n");
  NewLines(1);
  if( useAggregate )     WriteAll("FUNCTION - AGGREGATE\n");
		    else WriteAll("FUNCTION - SPLIT\n");
  if( useJacobian ) {
    if( useAllJacobian ) WriteAll("JACOBIAN - ALL\n");
		    else WriteAll("JACOBIAN - ON\n");
  } else {
    if( useAllJacobian ) WriteAll("JACOBIAN - Error\n");
		    else WriteAll("JACOBIAN - OFF\n");
  }
  if( useDouble )        WriteAll("DOUBLE   - ON\n");
		    else WriteAll("DOUBLE   - OFF\n");
  if( useReorder )        WriteAll("REORDER  - ON\n");
		    else WriteAll("REORDER  - OFF\n");
  NewLines(1);
  
  WriteAll("### Parameters ----------------------------------------\n");
  NewLines(1);

  DeclareConstant( NSPEC,   ascii( SpcNr ) );
  DeclareConstant( NVAR,    ascii( max( varNr, 1 ) ) );
  DeclareConstant( NVARACT, ascii( max( varactNr, 1 ) ) );
  DeclareConstant( NRAD,    ascii( max( radNr, 1 ) ) );
  DeclareConstant( NFIX,    ascii( max( fixNr, 1 ) ) );
  DeclareConstant( NREACT,  ascii( EqnNr ) );
  DeclareConstant( NVARST,  ascii( varstNr ) );
  DeclareConstant( NRADST,  ascii( radstNr ) );
  DeclareConstant( NFIXST,  ascii( fixstNr ) );

  NewLines(1);
  WriteAll("### Species -------------------------------------------\n");

  NewLines(1);
  WriteAll("Variable species\n");

  dn = varNr/3 + 1;
  for( i = 0; i < dn; i++ ) {
             if( i < varNr ) WriteSpec( i, i );
    i += dn; if( i < varNr ) WriteSpec( i, i ); 
    i += dn; if( i < varNr ) WriteSpec( i, i );
    i -= 2*dn; WriteAll("\n"); 
  }
    
  NewLines(1);
  WriteAll("Radical species\n");

  dn = radNr/3 + 1;
  for( i = 0; i < dn; i++ ) {
             if( i < radNr ) WriteSpec( i, i + varNr );
    i += dn; if( i < radNr ) WriteSpec( i, i + varNr ); 
    i += dn; if( i < radNr ) WriteSpec( i, i + varNr );
    i -= 2*dn; WriteAll("\n"); 
  }
    
  NewLines(1);
  WriteAll("Fixed species\n");

  dn = fixNr/3 + 1;
  for( i = 0; i < dn; i++ ) {
             if( i < fixNr ) WriteSpec( i, i + varNr + radNr );
    i += dn; if( i < fixNr ) WriteSpec( i, i + varNr + radNr ); 
    i += dn; if( i < fixNr ) WriteSpec( i, i + varNr + radNr );
    i -= 2*dn; WriteAll("\n"); 
  }
    
  NewLines(1);
  WriteAll("### Subroutines ---------------------------------------\n");
  NewLines(1);
  FlushBuf();
}

void GenerateInitVal()
{
int i;
int I, X;
int INITVAL;

  UseFile( srcFile );

  INITVAL    = DefFnc( "INITVAL",    0, "function to update initial concentrations");
  FunctionBegin( INITVAL );
  
  I = DefElm( "i", INT, 0);
  X = DefElm( "x", real, 0);
  Declare( I );
  Declare( X );

  NewLines(1);
  
  Assign( Elm( X ), Mul( Elm( IV, varDefault ), Elm( CFACTOR ) ) );
  C_Inline("  for( i = 0; i < NVAR; i++ )" );
  F_Inline("      do i = 1, NVAR" );
  F95_Inline("      do i = 1, NVAR" );				/* F95 line */
  ident++;
    Assign( Elm( VAR, -I ), Elm( X ) );
  ident--;
  F_Inline("      end do" );
  F95_Inline("      end do" );					/* F95 line */

  NewLines(1);
  Assign( Elm( X ), Mul( Elm( IV, radDefault ), Elm( CFACTOR ) ) );
  C_Inline("  for( i = 0; i < NRAD; i++ )" );
  F_Inline("      do i = 1, NRAD" );
  F95_Inline("      do i = 1, NRAD" );				/* F95 line */
  ident++;
    Assign( Elm( RAD, -I ), Elm( X ) );
  ident--;
  F_Inline("      end do" );
  F95_Inline("      end do" );					/* F95 line */

  NewLines(1);
  Assign( Elm( X ), Mul( Elm( IV, fixDefault ), Elm( CFACTOR ) ) );
  C_Inline("  for( i = 0; i < NFIX; i++ )" );
  F_Inline("      do i = 1, NFIX" );
  F95_Inline("      do i = 1, NFIX" );				/* F95 line */
  ident++;
    Assign( Elm( FIX, -I ), Elm( X ) );
  ident--;
  F_Inline("      end do" );
  F95_Inline("      end do" );					/* F95 line */


  NewLines(1);

  for( i = 0; i < varNr; i++) {
    if( *SpeciesTable[ Code[i] ].ival == 0 ) continue;
    Assign( Elm( VAR, i ), Mul( 
              Elm( IV, SpeciesTable[ Code[i] ].ival ),
              Elm( CFACTOR ) ) );
  }              

  for( i = 0; i < radNr; i++) {
    if( *SpeciesTable[ Code[i + varNr] ].ival == 0 ) continue;
    Assign( Elm( RAD, i ), Mul(
              Elm( IV, SpeciesTable[ Code[i + varNr] ].ival ),
              Elm( CFACTOR ) ) );  
  }
  
  for( i = 0; i < fixNr; i++) {
    if( *SpeciesTable[ Code[i + varNr + radNr] ].ival == 0 ) continue;
    Assign( Elm( FIX, i ), Mul(
              Elm( IV, SpeciesTable[ Code[i + varNr + radNr] ].ival ),
              Elm( CFACTOR ) ) ); 
  }             

  NewLines(1);
  C_Inline("  for( i = 0; i < NSPEC; i++ )" );
  F_Inline("      do i = 1, NSPEC" );
  F95_Inline("      do i = 1, NSPEC" );				/* F95 line */
  ident++;
    Assign( Elm( C_DEFAULT, -I ), Elm( C, -I ) );
  ident--;
  F_Inline("      end do" );
  F95_Inline("      end do" );					/* F95 line */

  NewLines(1);

  /* mz_rs_20020530+ */
  WriteComment("Set dummy default value for RCONST to avoid that the");
  WriteComment("non temperature-dependent elements remain undefined");
  F95_Inline("      RCONST(:)=-999." );
  NewLines(1);
  /* mz_rs_20020530- */

  WriteComment("User defined initialisations");
  
  switch( useLang ) {
    case C_LANG: bprintf( InlineCode[ C_INIT_INT ].code );
                 bprintf( InlineCode[ C_INIT ].code );
                 break;
    case F_LANG: bprintf( InlineCode[ F_INIT_INT ].code );
                 bprintf( InlineCode[ F_INIT ].code );
                 break;
    case F95_LANG: bprintf( InlineCode[ F95_INIT_INT ].code );	/* F95 line */
                 bprintf( InlineCode[ F95_INIT ].code );
                 break;
  }
  FlushBuf();

  NewLines(1);
  WriteComment("End user defined initialisations");
  NewLines(1);

  FreeVariable( X );
  FreeVariable( I );
  FunctionEnd( INITVAL );
  FreeVariable( INITVAL );
}

void GenerateUseConc()
{
int i,k,l;
int USECONC;

  UseFile( srcFile );

  USECONC    = DefFnc( "USECONC",    2, "function to copy concentrations from USER to KPP");
  FunctionBegin( USECONC, U_VAR, U_RAD );
  
  k = 0;l = 0;
  for( i = 1; i < SpcNr; i++) {
    if( ReverseCode[i] < 0 ) { 
      if( SpeciesTable[i].type == VAR_SPC ) k++; 
      continue;
    }  
    switch( SpeciesTable[i].type ) {
      case VAR_SPC: 
                     if( k < initNr ) {
                       Assign( Elm( C, ReverseCode[i] ), Elm( U_VAR, k++ ) ); 
                       break;
                     }  
      case RAD_SPC:  
                     Assign( Elm( C, ReverseCode[i] ), Elm( U_RAD, l++ ) ); 
                     break;
      case FIX_SPC:    
                     Assign( Elm( C, ReverseCode[i] ), Elm( C_DEFAULT, ReverseCode[i] ) );
                     break;
      case DUMMY_SPC:    
      default:       break;
    }
  }  

  FunctionEnd( USECONC );
  FreeVariable( USECONC );
}

void GenerateRestConc()
{
int i,k,l;
int RESTCONC;

  UseFile( srcFile );

  RESTCONC    = DefFnc( "RESTCONC",    2, "function to restore concentrations from KPP to USER");
  FunctionBegin( RESTCONC, U_VAR, U_RAD );
  
  k = 0; l = 0;
  for( i = 0; i < SpcNr; i++) {
    if( ReverseCode[i] < 0 ) { 
      if( SpeciesTable[i].type == VAR_SPC ) k++; 
      continue;
    }  
    switch( SpeciesTable[i].type ) {
      case VAR_SPC: 
                     if( k < initNr )
                       Assign( Elm( U_VAR, k++ ), Elm( C, ReverseCode[i] ) ); 
                     break;
      case RAD_SPC:  
                     Assign( Elm( U_RAD, l++ ), Elm( C, ReverseCode[i] ) ); 
                     break;
      case FIX_SPC:    
      case DUMMY_SPC:    
      default:       break;
    }
  }  

  FunctionEnd( RESTCONC );
  FreeVariable( RESTCONC );
}

void GenerateGetMass()
{
int i;
int atm, spc;
int GETMASS, MASS;
SPECIES_DEF *sp;
int nmass;

  UseFile( srcFile );

  nmass = 0;
  for( atm = 0; atm < AtomNr; atm++ )
    if( AtomTable[atm].masscheck ) nmass++;

  if( nmass == 0 ) nmass = 1;
  MASS  = DefvElm( "MASS", real, nmass, "value of mass balance" );
  GETMASS = DefFnc( "GETMASS", 1, "function that compute total mass of selected atoms");
  FunctionBegin( GETMASS, MASS );

  nmass = 0;
  for( atm = 0; atm < AtomNr; atm++ ) {
    if( AtomTable[atm].masscheck ) {
      sum = Const( 0 );
      for( spc = 0; spc < SpcNr; spc++ ) {
        sp = &SpeciesTable[ Code[spc] ]; 
        for( i = 0; i < sp->nratoms; i++ ) {
          if( sp->atoms[i].code == atm ) {
            sum = Add( sum, Mul( Const( sp->atoms[i].nr ), 
                                 Elm( C, spc ) ) );
          }
        }  
      }
      Assign( Elm( MASS, nmass ), sum );
      nmass++;
    }
  } 
  
  FunctionEnd( GETMASS );
  FreeVariable( GETMASS );
}

void GenerateMakefile()
{
FILE *fp;

  if( fileNr == 0 ) return;
  fp = fopen("makefile", "r");
  if( fp != 0 ) {
    fclose( fp );
    return;
  }
  fp = fopen("makefile", "w");
  if( fp == 0 ) {
    Warning("Can not create makefile");
    return;
  } 
  fprintf( fp, "cucu");
  fclose( fp ); 
}

void Generate()
{
int i, j;
int n;

  varstNr = 0;
  radstNr = varNr;
  fixstNr = varNr + radNr;
  
  real = useDouble ? DOUBLE : REAL;

  n = MAX_OUTBUF;
  for( i = 1; i < INLINE_OPT; i++ ) 
    if( InlineCode[i].maxlen > n ) 
      n = InlineCode[i].maxlen;
    
  outBuf = (char*)malloc( n );
  outBuffer = outBuf;

  switch( useLang ) {
    case F95_LANG: Use_F95( rootFileName ); 
                 break; 
    case F_LANG: Use_F( rootFileName ); 
                 break; 
    case C_LANG: Use_C( rootFileName ); 
                 break; 
  }
  InitGen();

  GenerateMap();
  GenerateGData();
  if( useJacSparse ) GenerateSparseData();
  /* GenerateUtil(); */        /* mz_rs_20020426: subroutine not used */

  if( useLang == F95_LANG ) { /* mz_rs_20040202 */
    F95_Inline("MODULE kpp");
    NewLines(1);
    F95_Inline("USE GDATA");
    NewLines(1);
    F95_Inline("IMPLICIT NONE");
    NewLines(1);
    F95_Inline("CONTAINS");
    NewLines(1);
  }
  GenerateFuncVar();
  GenerateFuncRad();
  if ( useJacobian ) {
    GenerateJacVar();
    GenerateJacRad();
    GenerateJVectVar();
    GenerateJVectRad();
    if( useJacSparse ) {
      GenerateSparseUtil();
      GenerateSolve();
    }  
  }
  GenerateInitVal();

  /* GenerateUseConc(); */     /* mz_rs_20020426: subroutine not used */
  /* GenerateRestConc(); */    /* mz_rs_20020426: subroutine not used */

  GenerateUpdateSun();
  GenerateUpdateRconst();
  GenerateUpdatePhoto();
  /* GenerateUpdateC(); */    /* mz_rs_20020426: subroutine not used */

  /* GenerateGetMass(); */    /* mz_rs_20020426: subroutine not used */

  GenerateIntegrate();
  if( useLang == F95_LANG ) { /* mz_rs_20040202 */
    F95_Inline( "END MODULE kpp" ); 			
  }
  GenerateDriver();


  if( useJacSparse ) {
    if( useLang == F95_LANG )
      GenerateSparseModuleHeader(); /* mz_rs_20020426: only for f95 */
    else
      GenerateSparseHeader(); /* mz_rs_20020426: not for f95 */
  } else {  /* create dummy sparse MODULE file mz_rs_20030530 */
    UseFile( sparseModuleFile );
    F95_Inline("MODULE SPRSDATA");
    NewLines(1);
    WriteComment("Dummy");
    NewLines(1);
    F95_Inline( "END MODULE SPRSDATA" ); 			
  }

  /* mz_rs_20020530: call GenerateHeader after GenerateSparseModuleHeader */
  /* because LU_nzv must already be defined */
  GenerateHeader();
  
  GenerateMakefile();

  /* if(   hFile ) fclose(   hFile ); */ /* mz_rs_20020426: not for f95 */
  if( srcFile ) fclose( srcFile );
  /* if( sparseFile ) fclose( sparseFile ); */ /* mz_rs_20020426: not for f95 */
  if( mapFile ) fclose( mapFile );
  if( sparseModuleFile ) fclose( sparseModuleFile );
  if( ModuleHeadFile ) fclose( ModuleHeadFile );
}
