/***********************************************************************************
  KPP - symbolic chemistry Kinetics PreProcessor

  developed at Center for Global and Regional Atmospheric Research,
               University of Iowa

            by Valeriu Damian & Adrian Sandu 

***********************************************************************************/

#include <stdlib.h>
#include <string.h>
#include "gdata.h"
#include "scan.h"

char *eqFileName;
char *rootFileName = "ff";
char Home[ MAX_PATH ] = ""; 

short int linStru[ MAX_SPECIES ];
short int colStru[ MAX_SPECIES ];
short int bestStru[ MAX_SPECIES ];
short int *Stru;

enum stru_criteria { UNSORT, LINSORT, COLSORT, BESTSORT };

void EqCopy( EQ_VECT e1, EQ_VECT e2 )
{
int i;

  for( i = 0; i < EqnNr; i++ ) e2[i] = e1[i];
}

int NoSort( const void *p1, const void *p2 )
{
  return -1;
}

int CodeCmp( const void *p1, const void *p2 )
{
CODE *c1, *c2;

  c1 = (CODE*)p1;  
  c2 = (CODE*)p2;  
  
  if ( *c1 < *c2 ) return -1;
  if ( *c1 > *c2 ) return 1;
  return 0;
}

int CodeRCmp( const void *p1, const void *p2 )
{
int rc1, rc2;
CODE *c1, *c2;

  c1 = (CODE*)p1;  
  c2 = (CODE*)p2;  

  rc1 = Reactive[ ReverseCode[ *c1 ] ]; 
  rc2 = Reactive[ ReverseCode[ *c2 ] ];
  if ( rc1 > rc2 ) return -1; 
  if ( rc1 < rc2 ) return 1;
  if ( *c1 < *c2 ) return -1;
  if ( *c1 > *c2 ) return 1;
  return 0;
}

int CodeSSCmp( const void *p1, const void *p2 )
{
  return -CodeRCmp(p1,p2);
}

int CodeSCmp( const void *p1, const void *p2 )
{
CODE *c1, *c2;
short int sc1, sc2;

  c1 = (CODE*)p1;  
  c2 = (CODE*)p2;  
  
  sc1 = Stru[ ReverseCode[ *c1 ] ];
  sc2 = Stru[ ReverseCode[ *c2 ] ];
  
  if ( sc1 > sc2 ) return 1; 
  if ( sc1 < sc2 ) return -1;
  if ( *c1 < *c2 ) return 1;
  if ( *c1 > *c2 ) return -1;
  return 0;
}

void UpdateStructJ()
{
int i,j,k;

  for ( i=0; i<varNr; i++ )
    for ( j=0; j<varNr; j++ )
      structJ[i][j]=(i==j)?1:0;
              
  for (i = 0; i < varNr; i++)
    for (j = 0; j < varNr; j++)
      for (k = 0; k < EqnNr; k++)
        if ( cf[i][k]*((cc[j][k])?1:0) != 0.0 )
          structJ[i][j]=1;

  for ( i=0; i<varNr; i++ )
    for ( j=0; j<varNr; j++ )
      LUstructJ[i][j]=structJ[i][j];
              
}

int ComputeLUStructJ()
{
int i,j,k;
int nu,nl;

  for (j = 0; j < varNr-1; j++) {
    for (i = j+1; i < varNr; i++) {
      if( LUstructJ[i][j] ) {
        for (k = j; k < varNr; k++) 
         /*   LUstructJ[i][k] += LUstructJ[j][k]; */
	   if ( LUstructJ[j][k] != 0 )
              LUstructJ[i][k] = 1; 
      }   
    }
  }  
  
  nu = 0; nl = 0;
  for (i = 0; i < varNr; i++) 
    for (j = 0; j < varNr; j++)
      if( LUstructJ[i][j] ) {
        if(i > j) nl++;
        if(i <= j) nu++;     
      }          

  return nu+nl;
}

int LUnonZero()
{
CODE v[MAX_SPECIES];
CODE *var;
int i,j,k;
int nu,nl;

  var = v;
  if( Stru != bestStru ) {
    for( i=0; i<varNr; i++ )
      var[i] = Code[i];
    qsort( (void*)var, varNr, sizeof(CODE), CodeSCmp );
  } else {
    var = bestStru;
  }
          
  for (i = 0; i < varNr; i++)
    for (j = 0; j < varNr; j++)
      LUstructJ[i][j] = structJ[ ReverseCode[var[i]] ][ ReverseCode[var[j]] ];  

  return ComputeLUStructJ();
}

void LinColSparsity()
{
int i,j,k;
int nlin, ncol;
FILE * fff;

  for ( i=0; i<varNr; i++ )
    for ( j=0; j<varNr; j++ )
      structJ[i][j]=(i==j)?1:0;
              
  for (i = 0; i < varNr; i++)
    for (j = 0; j < varNr; j++)
      for (k = 0; k < EqnNr; k++)
        if ( cf[i][k]*((cc[j][k])?1:0) != 0.0 )
          structJ[i][j]=1;

  for ( i=0; i<varNr; i++ ) {
    linStru[i] = 0;
    for (j = 0; j < varNr; j++)
      linStru[i] += structJ[i][j];
  }

  for ( i=0; i<varNr; i++ ) {
    colStru[i] = 0;
    for (j = 0; j < varNr; j++)
      colStru[i] += structJ[j][i];
    colStru[i] *= linStru[i];  
  }

  Stru = linStru;
  nlin = LUnonZero();
  Stru = colStru;
  ncol = LUnonZero();
  if( nlin <= ncol ) { 
    Stru = linStru;
    LUnonZero();
  } 
}

void BestSparsity()
{
int i,j,k;
int cnz, lnz;
int best, crt;
int best_i;
int tmp;
int s;

  UpdateStructJ();
      
  for ( i=0; i<varNr; i++ )
    bestStru[i] = Code[i];

  for ( s=0; s<varNr-1; s++ ) {
    best = MAX_SPECIES*MAX_SPECIES; best_i = 0;
    for ( i=s; i<varNr; i++ ) {
      cnz = 0;lnz = 0;
      for (j = s; j < varNr; j++) {
        cnz += (LUstructJ[i][j]?1:0);
        lnz += (LUstructJ[j][i]?1:0);
      }
      crt = (cnz-1)*(lnz-1);
      if( crt < best ) {
        best = crt;
        best_i = i;
      }    
    }
    for ( i=0; i<varNr; i++ ) {
      tmp = LUstructJ[s][i];
      LUstructJ[s][i] = LUstructJ[best_i][i];
      LUstructJ[best_i][i] = tmp;      
    }   
    for ( i=0; i<varNr; i++ ) {
      tmp = LUstructJ[i][s];
      LUstructJ[i][s] = LUstructJ[i][best_i];
      LUstructJ[i][best_i] = tmp;      
    }   
    tmp = bestStru[s];
    bestStru[s] = bestStru[best_i];
    bestStru[best_i] = tmp;
    
    for (i = s+1; i < varNr; i++) {
      if( LUstructJ[i][s] ) {
        for (k = s; k < varNr; k++) 
          LUstructJ[i][k] += LUstructJ[s][k];
      }   
    }
  }

  Stru = bestStru;
}

void ReorderSpecies( int criteria )
{
CODE *var;
CODE *rad;
CODE *fix;
CODE *dummy;
EQ_VECT *tmpcc;
EQ_VECT *tmpcd;
EQ_VECT *tmpcf;
CODE *tmpCode;
CODE *tmpReact;
int i, k;
int new;
int (*cmpVar)(const void *, const void *);
int (*cmpRad)(const void *, const void *);
int (*cmpFix)(const void *, const void *);
int dummyNr;

  cmpVar = CodeCmp;
  cmpRad = CodeCmp;
  cmpFix = CodeCmp;

  switch( criteria ) {
    case UNSORT:   cmpVar = useJacobian ? CodeRCmp : CodeCmp;
                   break;  
    case LINSORT:  cmpVar = useJacobian ? CodeSCmp : CodeCmp;
                   Stru = linStru;  
                   break;  
    case COLSORT:  cmpVar = useJacobian ? CodeSCmp : CodeCmp;
                   Stru = colStru;
                   break;  
    case BESTSORT: cmpVar = useJacobian ? NoSort : CodeCmp;
                   break;  
  }

  varNr = 0;
  varactNr = 0;
  radNr = 0;
  fixNr = 0;
  dummyNr = 0;

  var = (CODE*)malloc( SpcNr * sizeof(CODE) );
  rad = (CODE*)malloc( SpcNr * sizeof(CODE) );
  fix = (CODE*)malloc( SpcNr * sizeof(CODE) );
  dummy = (CODE*)malloc( 5 * sizeof(CODE) );
  tmpcc = (EQ_VECT*)malloc( SpcNr * sizeof(EQ_VECT) );
  tmpcd = (EQ_VECT*)malloc( SpcNr * sizeof(EQ_VECT) );
  tmpcf = (EQ_VECT*)malloc( SpcNr * sizeof(EQ_VECT) );
  tmpCode = (CODE*)malloc( SpcNr * sizeof(CODE) );
  tmpReact = (CODE*)malloc( SpcNr * sizeof(CODE) );

  for( i = 0; i < SpcNr; i++ ) {
    switch( SpeciesTable[ Code[i] ].type ) {
      case VAR_SPC:  var[ varNr++ ] = Code[ i ];
		     break;
      case RAD_SPC:  rad[ radNr++ ] = Code[ i ];
		     break;
      case FIX_SPC:  fix[ fixNr++ ] = Code[ i ]; 
		     break;
      case DUMMY_SPC:dummy[ dummyNr++ ] = Code[ i ]; 
		     break;
    }
  }

  if( Stru != bestStru ) {
    qsort( (void*)var, varNr, sizeof(CODE), cmpVar );
  } else {
    for( i = 0; i < SpcNr; i++ )
      var[i] = bestStru[i];
  }  
  qsort( (void*)rad, radNr, sizeof(CODE), cmpRad );
  qsort( (void*)fix, fixNr, sizeof(CODE), cmpFix );

  for( i = 0; i < SpcNr; i++ ) {
    EqCopy( cc[i], tmpcc[i] );
    EqCopy( cd[i], tmpcd[i] );
    EqCopy( cf[i], tmpcf[i] );
    tmpCode[i] = Code[i];
    tmpReact[i] = Reactive[i];
  }

  SpcNr -= dummyNr;
  dummyNr = 0;

  k = 0;
  for( i = 0; i < varNr; i++ ) {
    new = ReverseCode[ var[i] ];
    EqCopy( tmpcc[ new ], cc[ k ] );
    EqCopy( tmpcd[ new ], cd[ k ] );
    EqCopy( tmpcf[ new ], cf[ k ] );
    Code[ k ] = tmpCode[ new ];
    Reactive[ k ] = tmpReact[ new ];
    if( Reactive[ k ] ) varactNr++; 
    k++;
  }
  for( i = 0; i < radNr; i++ ) {
    new = ReverseCode[ rad[i] ];
    EqCopy( tmpcc[ new ], cc[ k ] );
    EqCopy( tmpcd[ new ], cd[ k ] );
    EqCopy( tmpcf[ new ], cf[ k ] );
    Code[ k ] = tmpCode[ new ];
    Reactive[ k ] = tmpReact[ new ];
    if( ! Reactive[ k ] ) Error("%s: Radical species MUST react", 
                                SpeciesTable[ Code[k] ].name);
    k++;
  }
  for( i = 0; i < fixNr; i++ ) {
    new = ReverseCode[ fix[i] ];
    EqCopy( tmpcc[ new ], cc[ k ] );
    EqCopy( tmpcd[ new ], cd[ k ] );
    EqCopy( tmpcf[ new ], cf[ k ] );
    Code[ k ] = tmpCode[ new ];
    Reactive[ k ] = tmpReact[ new ];
    k++;
  }
  for( i = 0; i < dummyNr; i++ ) {
    new = ReverseCode[ dummy[i] ];
    EqCopy( tmpcc[ new ], cc[ k ] );
    EqCopy( tmpcd[ new ], cd[ k ] );
    EqCopy( tmpcf[ new ], cf[ k ] );
    Code[ k ] = tmpCode[ new ];
    Reactive[ k ] = tmpReact[ new ];
    k++;
  }


  for( i = 0; i < SpcNr+dummyNr; i++ )
    ReverseCode[ Code[i] ] = i;

  free( tmpReact );
  free( tmpCode );
  free( tmpcf );
  free( tmpcd );
  free( tmpcc );
  free( dummy );
  free( fix );
  free( rad );
  free( var );   

  fflush(stdout);
}

/* Allocate Internal Arrays */
void  AllocInternalArrays( void )
{
int i;

if ( (cc =(float**)malloc(MAX_SPECIES*sizeof(float*)))==NULL ) {
    printf("\nFatal Error. Cannot allocate cc.\n");
    exit(-3);
    }
for (i=0; i<MAX_SPECIES; i++)    
    if ( (cc[i] = (float*)malloc(MAX_EQN*sizeof(float)))==NULL ) {
    printf("\nFatal Error. Cannot allocate cc[%d].\n",i);
    exit(-3);
    }

if ( (cd = (float**)malloc(MAX_SPECIES*sizeof(float*)))==NULL ) {
    printf("\nFatal Error. Cannot allocate cd.\n");
    exit(-3);
    }
for (i=0; i<MAX_SPECIES; i++)    
    if ( (cd[i] = (float*)malloc(MAX_EQN*sizeof(float)))==NULL ) {
    printf("\nFatal Error. Cannot allocate cd[%d].\n",i);
    exit(-3);
    }

if ( (cf = (float**)malloc(MAX_SPECIES*sizeof(float*)))==NULL ) {
    printf("\nFatal Error. Cannot allocate cf.\n");
    exit(-3);
    }
for (i=0; i<MAX_SPECIES; i++)    
    if ( (cf[i] = (float*)malloc(MAX_EQN*sizeof(float)))==NULL ) {
    printf("\nFatal Error. Cannot allocate cf[%d].\n",i);
    exit(-3);
    }
    

}


/* Allocate Structure Arrays */
void  AllocStructArrays( void )
{
int i;


if ( (structB = (int**)malloc(EqnNr*sizeof(int*)))==NULL ) {
    printf("\nFatal Error. Cannot allocate structB.\n");
    exit(-3);
    }
for (i=0; i<EqnNr; i++)    
    if ( (structB[i] =(int*) malloc(SpcNr*sizeof(int)))==NULL ) {
    printf("\nFatal Error. Cannot allocate structB[%d].\n",i);
    exit(-3);
    }
    
if ( (structJ = (int**)malloc(SpcNr*sizeof(int*)))==NULL ) {
    printf("\nFatal Error. Cannot allocate structJ.\n");
    exit(-3);
    }
for (i=0; i<SpcNr; i++)    
    if ( (structJ[i] =(int*) malloc(SpcNr*sizeof(int)))==NULL ) {
    printf("\nFatal Error. Cannot allocate structJ[%d].\n",i);
    exit(-3);
    }
    
if ( (LUstructJ = (int**)malloc(SpcNr*sizeof(int*)))==NULL ) {
    printf("\nFatal Error. Cannot allocate LUstructJ.\n");
    exit(-3);
    }
for (i=0; i<SpcNr; i++)    
    if ( (LUstructJ[i] = (int*)malloc(SpcNr*sizeof(int)))==NULL ) {
    printf("\nFatal Error. Cannot allocate LUstructJ[%d].\n",i);
    exit(-3);
    }

}

                    
int main( int argc, char * argv[] )
{
int status;
char name[ 200 ];
char *p;
int i,j;
  
  AllocInternalArrays();

  p = getenv("KPP_HOME");
  if( p ) strcpy( Home, p );

  switch( argc ) {
    case 3: eqFileName = argv[1];
            rootFileName = argv[2];
            break;
    case 2: eqFileName = argv[1];
            strcpy( name, eqFileName );
            p = name + strlen(name);
            while( p > name ) {
              if( *p == '.') { 
                *p = '\0';
                break;
              }
              p--;  
            } 
	    rootFileName = name;
	    break;
    default: FatalError(1,"\nUsage :"
		          "\n        kpp <equations file> [output file]\n");
  }

  status = ParseEquationFile( argv[1] );

  if( status ) FatalError(2,"%d errors and %d warnings encountered.", 
                           nError, nWarning ); 
  /* Allocate some internal data structures */
  AllocStructArrays();

  ReorderSpecies( UNSORT );
  if (useReorder==1){
    BestSparsity(); 
    ReorderSpecies( BESTSORT );
    }
  UpdateStructJ();
  ComputeLUStructJ();

  if( initNr == -1 ) initNr = varNr;

/*
  for(i=0; i<SpcNr; i++) {
    printf("%3d %3d %3d %10s   ", i, ReverseCode[i], Code[i], SpeciesTable[i].name );
    switch( SpeciesTable[i].type ) {
      case VAR_SPC:  printf("VARIABLE");break;
      case RAD_SPC:  printf("RADICAL");break;
      case FIX_SPC:  printf("FIXED");break;
      case DUMMY_SPC:printf("dummy");break;
      default:       printf("???");break;
    }     
    printf("\n");
  }
*/

  Generate( rootFileName );

  if( nError ) exit(4);
  if( nWarning ) exit(5);

  exit(0);
}
